from dataclasses import dataclass, field
from datetime import datetime
from typing import Any, cast

import pandas as pd
from cognite.client import data_modeling as dm
from rdflib import Namespace, URIRef

from cognite.neat._v0.core._constants import DEFAULT_NAMESPACE
from cognite.neat._v0.core._data_model.models._base_unverified import (
    UnverifiedComponent,
    UnverifiedDataModel,
)
from cognite.neat._v0.core._data_model.models.data_types import DataType
from cognite.neat._v0.core._data_model.models.entities import (
    ConceptEntity,
    MultiValueTypeInfo,
    UnknownEntity,
    load_value_type,
)
from cognite.neat._v0.core._utils.rdf_ import uri_display_name

from ._verified import (
    Concept,
    ConceptualDataModel,
    ConceptualMetadata,
    ConceptualProperty,
)


@dataclass
class UnverifiedConceptualMetadata(UnverifiedComponent[ConceptualMetadata]):
    space: str
    external_id: str
    version: str
    creator: str
    name: str | None = None
    description: str | None = None
    created: datetime | str | None = None
    updated: datetime | str | None = None
    physical: str | URIRef | None = None
    source_id: str | URIRef | None = None

    @classmethod
    def _get_verified_cls(cls) -> type[ConceptualMetadata]:
        return ConceptualMetadata

    def dump(self, **kwargs: Any) -> dict[str, Any]:
        output = super().dump()
        if self.created is None:
            output["created"] = datetime.now()
        if self.updated is None:
            output["updated"] = datetime.now()
        return output

    def as_data_model_id(self) -> dm.DataModelId:
        return dm.DataModelId(space=self.space, external_id=self.external_id, version=self.version)

    @property
    def prefix(self) -> str:
        return self.space

    @property
    def identifier(self) -> URIRef:
        """Globally unique identifier for the data model.

        !!! note
            Unlike namespace, the identifier does not end with "/" or "#".

        """
        return DEFAULT_NAMESPACE[f"data-model/unverified/conceptual/{self.space}/{self.external_id}/{self.version}"]

    @property
    def namespace(self) -> Namespace:
        """Namespace for the data model used for the entities in the data model."""
        return Namespace(f"{self.identifier}/")


@dataclass
class UnverifiedConceptualProperty(UnverifiedComponent[ConceptualProperty]):
    concept: ConceptEntity | str | UnknownEntity
    property_: str
    value_type: DataType | ConceptEntity | MultiValueTypeInfo | UnknownEntity | str
    name: str | None = None
    description: str | None = None
    min_count: int | None = None
    max_count: int | float | None = None
    default: Any | None = None
    instance_source: str | list[str] | None = None
    # Only used internally
    inherited: bool = False
    neatId: str | URIRef | None = None

    # linking
    physical: str | URIRef | None = None

    @classmethod
    def _get_verified_cls(cls) -> type[ConceptualProperty]:
        return ConceptualProperty

    def dump(self, default_prefix: str, **kwargs) -> dict[str, Any]:  # type: ignore
        output = super().dump()
        output["Concept"] = ConceptEntity.load(self.concept, prefix=default_prefix, return_on_failure=True)
        output["Value Type"] = load_value_type(self.value_type, default_prefix, return_on_failure=True)
        return output

    def copy(self, update: dict[str, Any], default_prefix: str) -> "UnverifiedConceptualProperty":
        return cast(
            UnverifiedConceptualProperty,
            type(self)._load({**self.dump(default_prefix), **update}),
        )


@dataclass
class UnverifiedConcept(UnverifiedComponent[Concept]):
    concept: ConceptEntity | str
    name: str | None = None
    description: str | None = None
    implements: str | list[ConceptEntity] | None = None
    instance_source: str | None = None
    neatId: str | URIRef | None = None
    # linking
    physical: str | URIRef | None = None

    @classmethod
    def _get_verified_cls(cls) -> type[Concept]:
        return Concept

    @property
    def concept_str(self) -> str:
        return str(self.concept)

    def dump(self, default_prefix: str, **kwargs) -> dict[str, Any]:  # type: ignore
        output = super().dump()
        parent: list[ConceptEntity] | None = None
        if isinstance(self.implements, str):
            self.implements = self.implements.strip()
            parent = [
                ConceptEntity.load(parent_str, prefix=default_prefix, return_on_failure=True)
                for parent_str in self.implements.split(",")
            ]
        elif isinstance(self.implements, list):
            parent = [
                ConceptEntity.load(parent_str, prefix=default_prefix, return_on_failure=True)
                for parent_str in self.implements
            ]
        output["Concept"] = ConceptEntity.load(self.concept, prefix=default_prefix, return_on_failure=True)
        output["Implements"] = parent
        return output


@dataclass
class UnverifiedConceptualDataModel(UnverifiedDataModel[ConceptualDataModel]):
    metadata: UnverifiedConceptualMetadata
    concepts: list[UnverifiedConcept] = field(default_factory=list)
    properties: list[UnverifiedConceptualProperty] = field(default_factory=list)
    prefixes: dict[str, Namespace] | None = None

    @classmethod
    def _get_verified_cls(cls) -> type[ConceptualDataModel]:
        return ConceptualDataModel

    def dump(self) -> dict[str, Any]:
        default_prefix = self.metadata.prefix

        return dict(
            Metadata=self.metadata.dump(),
            Properties=[prop.dump(default_prefix) for prop in self.properties],
            Concepts=[concept.dump(default_prefix) for concept in self.concepts],
            Prefixes=self.prefixes,
        )

    @classmethod
    def display_type_name(cls) -> str:
        return "UnverifiedInformationModel"

    @property
    def display_name(self) -> str:
        return uri_display_name(self.metadata.identifier)

    def _repr_html_(self) -> str:
        summary = {
            "type": "Logical Data Model",
            "intended for": "Information Architect",
            "name": self.metadata.name,
            "external_id": self.metadata.external_id,
            "space": self.metadata.space,
            "version": self.metadata.version,
            "concepts": len(self.concepts),
            "properties": len(self.properties),
        }

        return pd.DataFrame([summary]).T.rename(columns={0: ""})._repr_html_()  # type: ignore



from abc import ABC
from typing import Annotated, Any, Literal

from pydantic import Field, TypeAdapter, field_serializer
from pydantic_core.core_schema import FieldSerializationInfo

from cognite.neat._utils.useful_types import BaseModelObject

from ._references import ContainerReference
from ._types import Bool


class ConstraintDefinition(BaseModelObject, ABC):
    constraint_type: str


class UniquenessConstraintDefinition(ConstraintDefinition):
    constraint_type: Literal["uniqueness"] = "uniqueness"
    properties: list[str] = Field(
        description="List of properties included in the constraint.", min_length=1, max_length=10
    )
    by_space: Bool | None = Field(default=None, description="Whether to make the constraint space-specific.")


class RequiresConstraintDefinition(ConstraintDefinition):
    constraint_type: Literal["requires"] = "requires"
    require: ContainerReference = Field(description="Reference to an existing container.")

    @field_serializer("require", mode="plain")
    @classmethod
    def serialize_require(cls, require: ContainerReference, info: FieldSerializationInfo) -> dict[str, Any]:
        output = require.model_dump(**vars(info))
        output["type"] = "container"
        return output


Constraint = Annotated[
    UniquenessConstraintDefinition | RequiresConstraintDefinition,
    Field(discriminator="constraint_type"),
]

ConstraintAdapter: TypeAdapter[Constraint] = TypeAdapter(Constraint)



from collections.abc import Hashable, ItemsView, Iterator, KeysView, Mapping, ValuesView
from typing import TypeVar

from rdflib import URIRef

from cognite.neat._v0.core._utils.spreadsheet import SpreadsheetRead

T_Key = TypeVar("T_Key", bound=Hashable)
T_Value = TypeVar("T_Value")


class ImportContext(dict, Mapping[T_Key, T_Value]):
    def __init__(self, data: Mapping[T_Key, T_Value] | None = None) -> None:
        super().__init__(data or {})

    # The below methods are included to make better type hints in the IDE
    def __getitem__(self, k: T_Key) -> T_Value:
        return super().__getitem__(k)

    def __setitem__(self, k: T_Key, v: T_Value) -> None:
        super().__setitem__(k, v)

    def __delitem__(self, k: T_Key) -> None:
        super().__delitem__(k)

    def __iter__(self) -> Iterator[T_Key]:
        return super().__iter__()

    def keys(self) -> KeysView[T_Key]:  # type: ignore[override]
        return super().keys()

    def values(self) -> ValuesView[T_Value]:  # type: ignore[override]
        return super().values()

    def items(self) -> ItemsView[T_Key, T_Value]:  # type: ignore[override]
        return super().items()

    def get(self, __key: T_Key, __default: T_Value | None = None) -> T_Value:  # type: ignore[override]
        return super().get(__key, __default)

    def pop(self, __key: T_Key, __default: T_Value | None = None) -> T_Value:  # type: ignore[override]
        return super().pop(__key, __default)

    def popitem(self) -> tuple[T_Key, T_Value]:
        return super().popitem()


class SpreadsheetContext(ImportContext[str, SpreadsheetRead]):
    def __init__(self, data: dict[str, SpreadsheetRead] | None = None) -> None:
        """Initialize the SpreadsheetContext with a dictionary of SpreadsheetRead objects.

        Args:
            data (dict[str, SpreadsheetRead]): A dictionary where keys are sheet names and values are
                SpreadsheetRead objects containing the read data.
        """
        super().__init__(data or {})
        for k, v in self.items():
            if not isinstance(k, str):
                raise TypeError(f"Expected string key, got {type(k).__name__}")
            if not isinstance(v, SpreadsheetRead):
                raise TypeError(f"Expected SpreadsheetRead for key '{k}', got {type(v).__name__}")


class GraphContext(ImportContext[str, Mapping[URIRef, int]]):
    def __init__(self, data: dict[str, Mapping[URIRef, int]] | None = None) -> None:
        """Initialize the GraphContext with a dictionary of mappings from URIRef to int.

        Args:
            data (dict[str, Mapping[URIRef, int]]): A dictionary where keys are graph names and values are
                mappings of URIRef to integer identifiers.
        """
        super().__init__(data or {})
        for k, v in self.items():
            if not isinstance(k, str):
                raise TypeError(f"Expected string key, got {type(k).__name__}")
            if not isinstance(v, Mapping):
                raise TypeError(f"Expected Mapping[URIRef, int] for key '{k}', got {type(v).__name__}")
            for uri, value in v.items():
                if not isinstance(uri, URIRef):
                    raise TypeError(f"Expected URIRef key in mapping for '{k}', got {type(uri).__name__}")
                if not isinstance(value, int):
                    raise TypeError(f"Expected int value in mapping for '{k}', got {type(value).__name__}")



from functools import total_ordering
from typing import Any

from pydantic import BaseModel, Field, field_validator, model_serializer

from ._constants import (
    PREFIX_PATTERN,
    SUFFIX_PATTERN,
    VERSION_PATTERN,
    Undefined,
    Unknown,
    _UndefinedType,
    _UnknownType,
)


@total_ordering
class Entity(BaseModel, extra="ignore", populate_by_name=True):
    """Entity is a concept, class, property, datatype in semantics sense."""

    prefix: str | _UndefinedType = Field(default=Undefined, pattern=PREFIX_PATTERN, min_length=1, max_length=43)
    suffix: str = Field(min_length=1, max_length=255, pattern=SUFFIX_PATTERN)

    @model_serializer(when_used="unless-none", return_type=str)
    def as_str(self) -> str:
        return str(self)

    @field_validator("*", mode="before")
    def strip_string(cls, value: Any) -> Any:
        if isinstance(value, str):
            return value.strip()
        elif isinstance(value, list):
            return [entry.strip() if isinstance(entry, str) else entry for entry in value]
        return value

    def as_tuple(self) -> tuple[str, ...]:
        # We haver overwritten the serialization to str, so we need to do it manually
        extra: tuple[str, ...] = tuple(
            [
                str(v or "")
                for field_name in self.model_fields.keys()
                if (v := getattr(self, field_name)) and field_name not in {"prefix", "suffix"}
            ]
        )
        if isinstance(self.prefix, _UndefinedType):
            return str(self.suffix), *extra
        else:
            return self.prefix, str(self.suffix), *extra

    def __lt__(self, other: object) -> bool:
        if type(other) is not type(self):
            return NotImplemented
        return self.as_tuple() < other.as_tuple()  # type: ignore[attr-defined]

    def __eq__(self, other: object) -> bool:
        # We need to be explicit that we are not allowing comparison between different types
        if type(other) is not type(self):
            # requires explicit raising as NotImplemented would lead to running comparison
            raise TypeError(f"'==' not supported between instances of {type(self).__name__} and {type(other).__name__}")
        return self.as_tuple() == other.as_tuple()  # type: ignore[attr-defined]

    def __hash__(self) -> int:
        return hash(f"{type(self).__name__}({self})")

    def __str__(self) -> str:
        # there are three cases for string representation:
        # 1. only suffix is present -> return str(suffix)
        # 2. prefix and suffix are present -> return "prefix:suffix"
        # 3. prefix, suffix and other fields are present -> return "prefix:suffix(field1=value1,field2=value2)"

        model_dump = {k: v for k in self.model_fields if (v := getattr(self, k)) is not None}

        base_str = f"{self.prefix}:{self.suffix}" if not isinstance(self.prefix, _UndefinedType) else str(self.suffix)

        extra_fields = {
            (self.model_fields[k].alias or k): v for k, v in model_dump.items() if k not in {"prefix", "suffix"}
        }
        if extra_fields:
            extra_str = ",".join([f"{k}={v}" for k, v in extra_fields.items()])
            return f"{base_str}({extra_str})"
        else:
            return base_str

    def __repr__(self) -> str:
        # We have overwritten the serialization to str, so we need to do it manually
        model_dump = {k: v for k in self.model_fields if (v := getattr(self, k)) is not None}
        args = ",".join([f"{k}={v!r}" for k, v in model_dump.items()])
        return f"{type(self).__name__}({args})"


class ConceptEntity(Entity):
    version: str | None = Field(default=None, pattern=VERSION_PATTERN, max_length=43)


class UnknownEntity(ConceptEntity):
    prefix: _UndefinedType = Undefined
    suffix: _UnknownType = Unknown  # type: ignore[assignment]

    @property
    def id(self) -> str:
        return str(Unknown)



import itertools
from collections import UserList
from dataclasses import dataclass, field
from datetime import datetime
from typing import Any

from cognite.neat._data_model._fix import FixAction
from cognite.neat._data_model.deployer.data_classes import DeploymentResult
from cognite.neat._issues import ConsistencyError, IssueList, ModelSyntaxError
from cognite.neat._state_machine import State


@dataclass
class Change:
    agent: str
    activity: str
    source_state: State
    start: datetime
    end: datetime

    target_state: State | None = field(default=None)
    source_entity: str | None = field(default="ExternalEntity")
    target_entity: str | None = field(default="FailedEntity")
    issues: IssueList | None = field(default=None)
    errors: IssueList | None = field(default=None)
    fixes: list[FixAction] | None = field(default=None)
    # for time being setting to Any, can be refined later
    result: DeploymentResult | None = field(default=None)
    description: str | None = field(default=None)

    @staticmethod
    def standardize_activity_name(activity: str, start: datetime, end: datetime) -> str:
        """Create standardized activity name"""
        return f"{activity}_{start.timestamp()}-{end.timestamp()}"

    @property
    def successful(self) -> bool:
        """Check if change was successful"""
        return not self.errors

    @property
    def error_count(self) -> int:
        """Get number of errors"""
        return sum(
            1
            for issue in itertools.chain(self.issues or [], self.errors or [])
            if isinstance(issue, ModelSyntaxError | ConsistencyError)
        )

    def as_mixpanel_event(self) -> dict[str, Any]:
        """Convert change to mixpanel event format"""
        return {
            "agent": self.agent,
            "activity": self.activity,
            "sourceEntity": self.source_entity,
            "targetEntity": self.target_entity,
            "sourceState": type(self.source_state).__name__,
            "targetState": type(self.target_state).__name__ if self.target_state else "None",
            "duration_ms": int((self.end - self.start).total_seconds() * 1000),
            "successful": self.successful,
            "issues": [issue.code or "<no code>" for issue in self.issues] if self.issues else [],
            "errors": [error.code or "<no code>" for error in self.errors] if self.errors else [],
        }


class Provenance(UserList[Change]):
    def __delitem__(self, *args: Any, **kwargs: Any) -> None:
        raise TypeError("Cannot delete change from provenance")

    def __setitem__(self, *args: Any, **kwargs: Any) -> None:
        raise TypeError("Cannot modify change from provenance")

    @property
    def last_change(self) -> Change | None:
        return self[-1] if len(self) > 0 else None

    def can_agent_do_activity(self, activity: Any) -> bool:
        "Check if activity can be performed based on provenance"
        raise NotImplementedError("Not implemented yet")

    def provenance_without_failures(self) -> "Provenance":
        """This method removes all the failed changes from the provenance list."""
        raise NotImplementedError("Not implemented yet")

    def last_physical_data_model_read(self) -> Change | None:
        """Get last physical data model read change"""
        for change in reversed(self):
            if change.activity.startswith("to_data_model"):
                return change
        return None



from collections.abc import Sequence

from cognite.neat._data_model.models.dms import ContainerReference, ContainerRequest, ContainerResponse
from cognite.neat._utils.http_client import HTTPClient, SuccessResponse

from .api import Endpoint, NeatAPI
from .config import NeatClientConfig
from .data_classes import PagedResponse
from .filters import ContainerFilter


class ContainersAPI(NeatAPI):
    def __init__(self, neat_config: NeatClientConfig, http_client: HTTPClient) -> None:
        super().__init__(
            neat_config,
            http_client,
            endpoint_map={
                "apply": Endpoint("POST", "models/containers", item_limit=100),
                "retrieve": Endpoint("POST", "models/containers/byids", item_limit=100),
                "delete": Endpoint("POST", "models/containers/delete", item_limit=100),
                "list": Endpoint("GET", "models/containers", item_limit=1000),
            },
        )

    def _validate_page_response(self, response: SuccessResponse) -> PagedResponse[ContainerResponse]:
        return PagedResponse[ContainerResponse].model_validate_json(response.body)

    def _validate_id_response(self, response: SuccessResponse) -> list[ContainerReference]:
        return PagedResponse[ContainerReference].model_validate_json(response.body).items

    def apply(self, items: Sequence[ContainerRequest]) -> list[ContainerResponse]:
        """Apply (create or update) containers in CDF.

        Args:
            items: List of ContainerRequest objects to apply.
        Returns:
            List of ContainerResponse objects.
        """
        return self._request_item_response(items, "apply")

    def retrieve(self, items: list[ContainerReference]) -> list[ContainerResponse]:
        """Retrieve containers by their identifiers.

        Args:
            items: List of (space, external_id) tuples identifying the containers to retrieve.

        Returns:
            List of ContainerResponse objects.
        """
        return self._request_item_response(items, "retrieve")

    def delete(self, items: list[ContainerReference]) -> list[ContainerReference]:
        """Delete containers by their identifiers.

        Args:
            items: List of ContainerReference objects identifying the containers to delete.

        Returns:
            List of ContainerReference objects representing the deleted containers.
        """
        return self._request_id_response(items, "delete")

    def list(
        self, space: str | None = None, include_global: bool = False, limit: int | None = 10
    ) -> list[ContainerResponse]:
        """List containers in CDF Project.

        Args:
            space: If specified, only containers in this space are returned.
            include_global: If True, include global containers.
            limit: Maximum number of containers to return. If None, return all containers.

        Returns:
            List of ContainerResponse objects.
        """
        filter = ContainerFilter(space=space, include_global=include_global)
        return self._list(limit=limit, params=filter.dump())



from typing import Any

from cognite.neat._data_model.exporters import DMSExporter
from cognite.neat._data_model.importers import DMSImporter
from cognite.neat._data_model.transformers import Transformer

from ._base import State


class Undo:
    """
    Event to trigger undoing the last action.
    """

    pass


class ForbiddenState(State):
    """
    State representing forbidden transitions - returns to previous state.
    """

    def __init__(self, previous_state: State):
        self.previous_state = previous_state

    def transition(self, event: Any) -> State:
        # only "undo" to trigger going back to previous state
        if isinstance(event, Undo):
            return self.previous_state
        return self


class EmptyState(State):
    """
    The initial state with empty NEAT store.
    """

    def transition(self, event: Any | None) -> State:
        if isinstance(event, DMSImporter):
            return PhysicalState()
        return ForbiddenState(self)


class PhysicalState(State):
    """
    State with physical model loaded.
    """

    def transition(self, event: Any) -> State:
        if isinstance(event, DMSExporter | Transformer):
            return PhysicalState()

        return ForbiddenState(self)



from collections.abc import Mapping
from dataclasses import dataclass, field

from .data_classes import DMS_API_MAPPING


@dataclass
class SpreadsheetReadContext:
    """This class is used to store information about the source spreadsheet.

    It is used to adjust row numbers to account for header rows and empty rows
    such that the error/warning messages are accurate.
    """

    header_row: int = 0
    empty_rows: list[int] = field(default_factory=list)

    def __post_init__(self) -> None:
        self.empty_rows.sort()

    def adjusted_row_number(self, row_no: int) -> int:
        """Adjusts the given row number to account for header rows and empty rows.

        Args:
            row_no (int): The original row number (0-based).

        !!! note "Row Numbering"
            Input rows are zero-indexed, while output rows are one-indexed as they appear in Excel.
            Therefore, we are adding 1 to offset header and 1 to offset row no

        """
        output = (row_no + 1) + (self.header_row + 1)
        counter = 0
        for empty_row in self.empty_rows:
            if empty_row < self.header_row:
                continue
            if empty_row <= output:
                counter += 1
                output += 1
            else:
                break

        return output


@dataclass
class TableSource:
    source: str
    table_read: dict[str, SpreadsheetReadContext] = field(default_factory=dict)

    def location(self, path: tuple[int | str, ...]) -> str:
        table_id: str | None = None
        row_no: int | None = None
        column: str | None = None
        if len(path) >= 1 and isinstance(path[0], str):
            table_id = path[0]
        if len(path) >= 2 and isinstance(path[1], int):
            row_no = path[1]
        if len(path) >= 3 and isinstance(path[2], str):
            column = path[2]
            column = self.field_to_column(table_id, column)
        if isinstance(row_no, int):
            row_no = self.adjust_row_number(table_id, row_no)
        location_parts = []
        if table_id is not None:
            location_parts.append(f"table {table_id!r}")
        if row_no is not None:
            location_parts.append(f"row {row_no}")
        if column is not None:
            location_parts.append(f"column {column!r}")
        if len(path) >= 4:
            location_parts.append("-> " + ".".join(str(p) for p in path[3:]))

        return " ".join(location_parts)

    def adjust_row_number(self, table_id: str | None, row_no: int) -> int:
        table_read = table_id and self.table_read.get(table_id)
        if table_read:
            return table_read.adjusted_row_number(row_no)
        return row_no + 1  # Convert to 1-indexed if no table read info is available

    @classmethod
    def field_to_column(cls, table_id: str | None, field_: str) -> str:
        """Maps the field name used in the DMS API to the column named used by Neat."""
        mapping = cls.field_mapping(table_id)
        if mapping is not None:
            return mapping.get(field_, field_)
        return field_

    @classmethod
    def field_mapping(cls, table_id: str | int | None) -> Mapping[str, str] | None:
        if not isinstance(table_id, str):
            return None
        return DMS_API_MAPPING.get(table_id)



from abc import ABC, abstractmethod
from collections.abc import Iterable, Sequence
from dataclasses import dataclass
from typing import Any, Generic, Literal, TypeAlias, TypeVar

from pydantic import BaseModel, JsonValue, TypeAdapter

from cognite.neat._client.config import NeatClientConfig
from cognite.neat._data_model.models.dms._base import T_Resource, T_Response
from cognite.neat._utils.collection import chunker_sequence
from cognite.neat._utils.http_client import HTTPClient, ParametersRequest, SimpleBodyRequest, SuccessResponse
from cognite.neat._utils.useful_types import T_Reference

from .data_classes import PagedResponse

_T_BaseModel = TypeVar("_T_BaseModel", bound=BaseModel)


@dataclass(frozen=True)
class Endpoint:
    method: Literal["GET", "POST"]
    path: str
    item_limit: int


APIMethod: TypeAlias = Literal["apply", "retrieve", "delete", "list"]


class NeatAPI(Generic[T_Reference, T_Resource, T_Response], ABC):
    def __init__(
        self, neat_config: NeatClientConfig, http_client: HTTPClient, endpoint_map: dict[APIMethod, Endpoint]
    ) -> None:
        self._config = neat_config
        self._http_client = http_client
        self._method_endpoint_map = endpoint_map

    @abstractmethod
    def _validate_page_response(self, response: SuccessResponse) -> PagedResponse[T_Response]:
        """Parse a single item response."""
        raise NotImplementedError()

    @abstractmethod
    def _validate_id_response(self, response: SuccessResponse) -> list[T_Reference]:
        """Parse a single item response."""
        raise NotImplementedError()

    def _make_url(self, path: str = "") -> str:
        """Create the full URL for this resource endpoint."""
        return self._config.create_api_url(path)

    def _request_item_response(
        self,
        items: Sequence[BaseModel],
        method: APIMethod,
        extra_body: dict[str, Any] | None = None,
    ) -> list[T_Response]:
        response_items: list[T_Response] = []
        for response in self._chunk_requests(items, method, extra_body):
            response_items.extend(self._validate_page_response(response).items)
        return response_items

    def _request_id_response(
        self,
        items: Sequence[BaseModel],
        method: APIMethod,
        extra_body: dict[str, Any] | None = None,
    ) -> list[T_Reference]:
        response_items: list[T_Reference] = []
        for response in self._chunk_requests(items, method, extra_body):
            response_items.extend(self._validate_id_response(response))
        return response_items

    def _chunk_requests(
        self,
        items: Sequence[_T_BaseModel],
        method: APIMethod,
        extra_body: dict[str, Any] | None = None,
    ) -> Iterable[SuccessResponse]:
        """Send requests in chunks and yield responses.

        Args:
            items: The items to process.
            method: The API method to use. This is used ot look the up the endpoint.
            extra_body: Optional extra body content to include in the request.

        Yields:
            The successful responses from the API.
        """
        # Filter out None
        endpoint = self._method_endpoint_map[method]

        for chunk in chunker_sequence(items, endpoint.item_limit):
            request = SimpleBodyRequest(
                endpoint_url=self._make_url(endpoint.path),
                method=endpoint.method,
                body=TypeAdapter(dict[str, JsonValue]).dump_json(
                    {
                        "items": [item.model_dump(by_alias=True) for item in chunk],
                        **(extra_body or {}),
                    },
                ),
            )
            response = self._http_client.request_with_retries(request)
            response.raise_for_status()
            yield response.success_response

    @classmethod
    def _filter_out_none_values(cls, params: dict[str, Any] | None) -> dict[str, Any] | None:
        request_params: dict[str, Any] | None = None
        if params:
            request_params = {k: v for k, v in params.items() if v is not None}
        return request_params

    def _paginate(
        self,
        limit: int,
        cursor: str | None = None,
        params: dict[str, Any] | None = None,
    ) -> PagedResponse[T_Response]:
        """Fetch a single page of resources.

        Args:
            params: Query parameters for the request. Supported parameters depend on
                the resource type but typically include:
                - cursor: Cursor for pagination
                - limit: Maximum number of items (defaults to list limit)
                - space: Filter by space
                - includeGlobal: Whether to include global resources
            limit: Maximum number of items to return in the page.
            cursor: Cursor for pagination.

        Returns:
            A Page containing the items and the cursor for the next page.
        """
        endpoint = self._method_endpoint_map["list"]
        if not (0 < limit <= endpoint.item_limit):
            raise ValueError(f"Limit must be between 1 and {endpoint.item_limit}, got {limit}.")
        if endpoint.method != "GET":
            raise NotImplementedError(f"Pagination not implemented for method {endpoint.method}.")
        request_params = self._filter_out_none_values(params) or {}
        request_params["limit"] = limit
        if cursor is not None:
            request_params["cursor"] = cursor
        request = ParametersRequest(
            endpoint_url=self._make_url(endpoint.path),
            method=endpoint.method,
            parameters=request_params,
        )
        result = self._http_client.request_with_retries(request)
        result.raise_for_status()
        return self._validate_page_response(result.success_response)

    def _iterate(
        self,
        limit: int | None = None,
        cursor: str | None = None,
        params: dict[str, Any] | None = None,
    ) -> Iterable[list[T_Response]]:
        """Iterate over all resources, handling pagination automatically."""
        next_cursor = cursor
        total = 0
        endpoint = self._method_endpoint_map["list"]
        while True:
            page_limit = endpoint.item_limit if limit is None else min(limit - total, endpoint.item_limit)
            page = self._paginate(limit=page_limit, cursor=next_cursor, params=params)
            yield page.items
            total += len(page.items)
            if page.next_cursor is None or (limit is not None and total >= limit):
                break
            next_cursor = page.next_cursor

    def _list(
        self,
        limit: int | None = None,
        params: dict[str, Any] | None = None,
    ) -> list[T_Response]:
        """List all resources, handling pagination automatically."""
        return [item for batch in self._iterate(limit=limit, params=params) for item in batch]



from abc import ABC, abstractmethod
from typing import ClassVar

from cognite.neat._data_model._analysis import ValidationResources
from cognite.neat._data_model._fix import FixAction
from cognite.neat._issues import ConsistencyError, Recommendation


class NeatRule(ABC):
    """Rules for data model principles."""

    code: ClassVar[str]
    issue_type: ClassVar[type[ConsistencyError] | type[Recommendation]]
    alpha: ClassVar[bool] = False
    fixable: ClassVar[bool] = False

    def __init__(
        self,
        validation_resources: ValidationResources,
    ) -> None:
        self.validation_resources = validation_resources

    @abstractmethod
    def validate(self) -> list[ConsistencyError] | list[Recommendation] | list[ConsistencyError | Recommendation]:
        """Execute rule validation."""
        ...

    def fix(self) -> list[FixAction]:
        """Fix the issues found by the validator producing a fixed object."""

        raise NotImplementedError("This rule does not implement fix()")



from collections import defaultdict

from cognite.neat._data_model._fix import FixAction
from cognite.neat._data_model.deployer.data_classes import (
    AddedField,
    ChangedField,
    FieldChange,
    PrimitiveField,
    RemovedField,
)
from cognite.neat._data_model.models.dms import (
    ContainerReference,
    DataModelReference,
    DataModelResource,
    SchemaResourceId,
    SpaceReference,
    ViewReference,
)
from cognite.neat._data_model.models.dms._schema import RequestSchema
from cognite.neat._data_model.transformers._base import Transformer


class FixApplicator(Transformer):
    """Applies the changes in FixAction objects to a schema."""

    def __init__(self, fix_actions: list[FixAction]) -> None:
        self._fix_actions = fix_actions

    def transform(self, data_model: RequestSchema) -> RequestSchema:
        """Apply fix actions and return the fixed schema (a deep copy of the original)."""
        result = data_model.model_copy(deep=True)

        if not self._fix_actions:
            return result

        fix_by_resource_id: dict[SchemaResourceId, list[FixAction]] = defaultdict(list)
        for action in self._fix_actions:
            fix_by_resource_id[action.resource_id].append(action)

        resources_list_lookup: dict[type, dict[SchemaResourceId, DataModelResource]] = {
            ViewReference: {view.as_reference(): view for view in result.views},
            ContainerReference: {container.as_reference(): container for container in result.containers},
            SpaceReference: {space.as_reference(): space for space in result.spaces},
            DataModelReference: {result.data_model.as_reference(): result.data_model},
        }

        for resource_id, actions in fix_by_resource_id.items():
            resource_lookup = resources_list_lookup.get(type(resource_id))
            if resource_lookup is None:
                raise RuntimeError(
                    f"{type(self).__name__}: Unsupported resource type {type(resource_id)}. This is a bug in NEAT."
                )
            resource = resource_lookup.get(resource_id)
            if resource is None:
                raise RuntimeError(
                    f"{type(self).__name__}: Resource {resource_id} not found in schema. This is a bug in NEAT."
                )

            all_changes_for_resource = [change for action in actions for change in action.changes]
            self._check_no_field_path_conflicts(all_changes_for_resource)
            self._apply_changes_to_resource(resource, all_changes_for_resource)

        return result

    def _apply_changes_to_resource(self, resource: DataModelResource, changes: list[FieldChange]) -> None:
        """Apply field changes to the resource in place."""
        for change in changes:
            if not isinstance(change, PrimitiveField):
                raise RuntimeError(
                    f"{type(self).__name__}: Only primitive field changes are supported, "
                    f"got {type(change).__name__}. This is a bug in NEAT."
                )
            if "." not in change.field_path:
                raise RuntimeError(
                    f"{type(self).__name__}: Invalid field_path '{change.field_path}' "
                    "(expected 'field_name.identifier' format). This is a bug in NEAT."
                )
            field_name, identifier = change.field_path.split(".", maxsplit=1)
            field_map = getattr(resource, field_name, None)
            if field_map is None:
                field_map = {}
                setattr(resource, field_name, field_map)
            if isinstance(change, RemovedField):
                field_map.pop(identifier, None)
            elif isinstance(change, AddedField | ChangedField):
                field_map[identifier] = change.new_value
            if not field_map:
                setattr(resource, field_name, None)

    def _check_no_field_path_conflicts(self, changes: list[FieldChange]) -> None:
        """Raise if any changes touch a field_path already modified by a previous change."""
        seen_paths: set[str] = set()
        for change in changes:
            if change.field_path in seen_paths:
                raise RuntimeError(
                    f"{type(self).__name__}: Conflicting fixes — multiple changes "
                    f"to '{change.field_path}'. This is a bug in NEAT."
                )
            seen_paths.add(change.field_path)



SPACE_FORMAT_PATTERN = r"^[a-zA-Z][a-zA-Z0-9_-]{0,41}[a-zA-Z0-9]?$"
DM_EXTERNAL_ID_PATTERN = r"^[a-zA-Z]([a-zA-Z0-9_]{0,253}[a-zA-Z0-9])?$"
CONTAINER_AND_VIEW_PROPERTIES_IDENTIFIER_PATTERN = r"^[a-zA-Z0-9][a-zA-Z0-9_-]{0,253}[a-zA-Z0-9]?$"
INSTANCE_ID_PATTERN = r"^[^\x00]{1,256}$"
ENUM_VALUE_IDENTIFIER_PATTERN = r"^[_A-Za-z][_0-9A-Za-z]{0,127}$"
DM_VERSION_PATTERN = r"^[a-zA-Z0-9]([.a-zA-Z0-9_-]{0,41}[a-zA-Z0-9])?$"
DATA_MODEL_DESCRIPTION_MAX_LENGTH = 1024
FORBIDDEN_ENUM_VALUES = frozenset({"true", "false", "null"})
FORBIDDEN_SPACES = frozenset(["space", "cdf", "dms", "pg3", "shared", "system", "node", "edge"])
FORBIDDEN_CONTAINER_AND_VIEW_EXTERNAL_IDS = frozenset(
    [
        "Query",
        "Mutation",
        "Subscription",
        "String",
        "Int32",
        "Int64",
        "Int",
        "Float32",
        "Float64",
        "Float",
        "Timestamp",
        "JSONObject",
        "Date",
        "Numeric",
        "Boolean",
        "PageInfo",
        "File",
        "Sequence",
        "TimeSeries",
    ]
)
FORBIDDEN_CONTAINER_AND_VIEW_PROPERTIES_IDENTIFIER = frozenset(
    [
        "space",
        "externalId",
        "createdTime",
        "lastUpdatedTime",
        "deletedTime",
        "edge_id",
        "node_id",
        "project_id",
        "property_group",
        "seq",
        "tg_table_name",
        "extensions",
    ]
)



"""This module performs importing of data model from spreadsheets."""

import tempfile
from collections import UserDict, defaultdict
from dataclasses import dataclass
from pathlib import Path
from typing import Literal, cast

import pandas as pd
from openpyxl import load_workbook
from openpyxl.worksheet.worksheet import Worksheet
from pandas import ExcelFile
from pydantic import ValidationError
from rdflib import Namespace, URIRef

from cognite.neat._v0.core._data_model._constants import SPLIT_ON_COMMA_PATTERN
from cognite.neat._v0.core._data_model._shared import (
    ImportedDataModel,
    T_UnverifiedDataModel,
)
from cognite.neat._v0.core._data_model.models import (
    UNVERIFIED_DATA_MODEL_BY_ROLE,
    VERIFIED_DATA_MODEL_BY_ROLE,
    RoleTypes,
    SchemaCompleteness,
)
from cognite.neat._v0.core._data_model.models._import_contexts import SpreadsheetContext
from cognite.neat._v0.core._data_model.models.entities._single_value import ContainerConstraintEntity, ContainerEntity
from cognite.neat._v0.core._issues import IssueList, MultiValueError
from cognite.neat._v0.core._issues.errors import (
    FileMissingRequiredFieldError,
    FileNotFoundNeatError,
    FileReadError,
)
from cognite.neat._v0.core._issues.warnings import FileMissingRequiredFieldWarning
from cognite.neat._v0.core._utils.spreadsheet import (
    SpreadsheetRead,
    find_column_and_row_with_value,
    read_individual_sheet,
)
from cognite.neat._v0.core._utils.text import humanize_collection

from ._base import BaseImporter

SOURCE_SHEET__TARGET_FIELD__HEADERS = [
    (
        "Properties",
        "Properties",
        {
            RoleTypes.information: ["Concept", "Property"],
            RoleTypes.dms: ["View", "View Property"],
        },
    ),
    ("Concepts", "Concepts", ["Concept"]),
    ("Containers", "Containers", ["Container"]),
    ("Views", "Views", ["View"]),
    ("Enum", "Enum", ["Collection"]),
    ("Nodes", "Nodes", ["Node"]),
]


MANDATORY_SHEETS_BY_ROLE: dict[RoleTypes, set[str]] = {
    role_type: {
        str(sheet_name)
        for sheet_name in (
            VERIFIED_DATA_MODEL_BY_ROLE.get(role_type).mandatory_fields(use_alias=True)  # type: ignore
            if VERIFIED_DATA_MODEL_BY_ROLE.get(role_type)
            else []
        )
        if sheet_name is not None
    }
    for role_type in RoleTypes.__members__.values()
}


class MetadataRaw(UserDict):
    @classmethod
    def from_excel(cls, excel_file: ExcelFile, metadata_sheet_name: str) -> "MetadataRaw":
        return cls(pd.read_excel(excel_file, metadata_sheet_name, header=None).replace(float("nan"), None).values)

    @property
    def has_role_field(self) -> bool:
        return self.get("role") in [role.value for role in RoleTypes.__members__.values()]

    @property
    def role(self) -> RoleTypes:
        return RoleTypes(self["role"])

    @property
    def has_schema_field(self) -> bool:
        return self.get("schema") in [schema.value for schema in SchemaCompleteness.__members__.values()]

    @property
    def schema(self) -> SchemaCompleteness | None:
        if not self.has_schema_field:
            return None
        return SchemaCompleteness(self["schema"])

    def is_valid(self, issue_list: IssueList, filepath: Path) -> bool:
        if not self.has_role_field:
            issue_list.append(FileMissingRequiredFieldError(filepath, "metadata", "role"))
            return False

        return True


@dataclass
class ReadResult:
    sheets: dict[str, dict | list]
    read_info_by_sheet: dict[str, SpreadsheetRead]
    metadata: MetadataRaw
    prefixes: dict[str, Namespace] | None = None

    @property
    def role(self) -> RoleTypes:
        return self.metadata.role

    @property
    def schema(self) -> SchemaCompleteness | None:
        return self.metadata.schema


class SpreadsheetReader:
    def __init__(
        self,
        issue_list: IssueList,
        required: bool = True,
        metadata: MetadataRaw | None = None,
        sheet_prefix: Literal["", "Last", "Ref", "CDMRef"] = "",
    ):
        self.issue_list = issue_list
        self.required = required
        self.metadata = metadata
        self._sheet_prefix = sheet_prefix
        self._seen_files: set[Path] = set()
        self._seen_sheets: set[str] = set()

    @property
    def metadata_sheet_name(self) -> str:
        return f"{self._sheet_prefix}Metadata"

    @property
    def prefixes_sheet_name(self) -> str:
        return "Prefixes"

    @property
    def seen_sheets(self) -> set[str]:
        if not self._seen_files:
            raise ValueError("No files have been read yet.")
        return self._seen_sheets

    def sheet_names(self, role: RoleTypes) -> set[str]:
        names = MANDATORY_SHEETS_BY_ROLE[role]
        return {f"{self._sheet_prefix}{sheet_name}" for sheet_name in names if sheet_name != "Metadata"}

    def read(self, excel_file: pd.ExcelFile, filepath: Path) -> None | ReadResult:
        self._seen_files.add(filepath)
        self._seen_sheets.update(map(str, excel_file.sheet_names))
        metadata: MetadataRaw | None
        if self.metadata is not None:
            metadata = self.metadata
        else:
            metadata = self._read_metadata(excel_file, filepath)
            if metadata is None:
                # The reading of metadata failed, so we can't continue
                return None

        sheets, read_info_by_sheet = self._read_sheets(excel_file, metadata.role)
        if sheets is None or self.issue_list.has_errors:
            return None
        sheets["Metadata"] = dict(metadata)

        # Special case for reading prefixes as they are suppose to be read only once
        if (
            self.prefixes_sheet_name in excel_file.sheet_names
            and not self._sheet_prefix
            and (prefixes := self._read_prefixes(excel_file, filepath))
        ):
            sheets["Prefixes"] = prefixes

        return ReadResult(sheets, read_info_by_sheet, metadata)

    def _read_metadata(self, excel_file: ExcelFile, filepath: Path) -> MetadataRaw | None:
        if self.metadata_sheet_name not in excel_file.sheet_names:
            if self.required:
                self.issue_list.append(FileMissingRequiredFieldError(filepath, "sheet", self.metadata_sheet_name))
            return None

        metadata = MetadataRaw.from_excel(excel_file, self.metadata_sheet_name)

        if not metadata.is_valid(self.issue_list, filepath):
            return None
        return metadata

    def _read_prefixes(self, excel_file: ExcelFile, filepath: Path) -> dict[str, Namespace] | None:
        if self.prefixes_sheet_name not in excel_file.sheet_names:
            return None

        else:
            prefixes = {}

            for row in read_individual_sheet(excel_file, "Prefixes", expected_headers=["Prefix", "Namespace"]):
                if "Prefix" in row and "Namespace" in row:
                    prefixes[row["Prefix"]] = row["Namespace"]
                else:
                    if "Prefix" not in row:
                        self.issue_list.append(FileMissingRequiredFieldWarning(filepath, "prefixes", "prefix"))
                    if "Namespace" not in row:
                        self.issue_list.append(FileMissingRequiredFieldWarning(filepath, "prefixes", "namespace"))
                    return None

            return prefixes

    def _read_sheets(
        self, excel_file: ExcelFile, read_role: RoleTypes
    ) -> tuple[dict[str, dict | list] | None, dict[str, SpreadsheetRead]]:
        read_info_by_sheet: dict[str, SpreadsheetRead] = defaultdict(SpreadsheetRead)

        sheets: dict[str, dict | list] = {}

        expected_sheet_names = self.sheet_names(read_role)

        if missing_sheets := expected_sheet_names.difference(set(excel_file.sheet_names)):
            if self.required:
                self.issue_list.append(
                    FileMissingRequiredFieldError(
                        cast(Path, excel_file.io), "sheets", humanize_collection(missing_sheets)
                    )
                )
            return None, read_info_by_sheet

        for source_sheet_name, target_sheet_name, headers_input in SOURCE_SHEET__TARGET_FIELD__HEADERS:
            source_sheet_name = f"{self._sheet_prefix}{source_sheet_name}"

            if source_sheet_name not in excel_file.sheet_names:
                continue
            if isinstance(headers_input, dict):
                headers = headers_input[read_role]
            else:
                headers = headers_input

            try:
                sheets[target_sheet_name], read_info_by_sheet[source_sheet_name] = read_individual_sheet(
                    excel_file,
                    source_sheet_name,
                    return_read_info=True,
                    expected_headers=headers,
                )
            except Exception as e:
                self.issue_list.append(FileReadError(cast(Path, excel_file.io), str(e)))
                continue

        return sheets, read_info_by_sheet


class ExcelImporter(BaseImporter[T_UnverifiedDataModel]):
    """Import data_model from an Excel file.

    Args:
        filepath (Path): The path to the Excel file.
    """

    def __init__(self, filepath: Path):
        self.filepath = filepath

    def to_data_model(self) -> ImportedDataModel[T_UnverifiedDataModel]:
        issue_list = IssueList(title=f"'{self.filepath.name}'")
        if not self.filepath.exists():
            raise FileNotFoundNeatError(self.filepath)

        self.filepath = self._make_forward_compatible_spreadsheet(self.filepath)

        with pd.ExcelFile(self.filepath) as excel_file:
            user_reader = SpreadsheetReader(issue_list)
            user_read = user_reader.read(excel_file, self.filepath)

        issue_list.trigger_warnings()
        if issue_list.has_errors:
            raise MultiValueError(issue_list.errors)

        if user_read is None:
            return ImportedDataModel(None, None)

        sheets = user_read.sheets
        original_role = user_read.role
        read_info_by_sheet = user_read.read_info_by_sheet

        data_model_cls = UNVERIFIED_DATA_MODEL_BY_ROLE[original_role]
        data_model = cast(T_UnverifiedDataModel, data_model_cls.load(sheets))

        # Delete the temporary file if it was created
        if "temp_neat_file" in self.filepath.name:
            try:
                self.filepath.unlink()
            except Exception as e:
                issue_list.append(FileReadError(self.filepath, f"Failed to delete temporary file: {e}"))

        return ImportedDataModel(data_model, SpreadsheetContext(read_info_by_sheet))

    @property
    def description(self) -> str:
        return f"Excel file {self.filepath.name} read as unverified data model"

    @property
    def source_uri(self) -> URIRef:
        return URIRef(f"file://{self.filepath.name}")

    def _make_forward_compatible_spreadsheet(self, filepath: Path) -> Path:
        """Makes the spreadsheet forward compatible by renaming legacy class with concept

        Args:
            filepath (Path): The path to the Excel file.

        """

        workbook = load_workbook(filepath, data_only=True)

        if "Classes" in workbook.sheetnames:
            print(
                (
                    "You are using a legacy spreadsheet format, "
                    "which we will support until v1.0 of neat."
                    " Please update your spreadsheet to the new format."
                ),
            )
            _replace_class_with_concept_cell(workbook["Classes"])
            sheet = workbook["Classes"]
            sheet.title = "Concepts"

            if "Properties" in workbook.sheetnames:
                _replace_class_with_concept_cell(workbook["Properties"])

        elif "Containers" in workbook.sheetnames:
            _replace_legacy_constraint_form(workbook["Containers"])
            _replace_legacy_constraint_form(workbook["Properties"])

        else:
            return filepath

        with tempfile.NamedTemporaryFile(prefix="temp_neat_file", suffix=".xlsx", delete=False) as temp_file:
            workbook.save(temp_file.name)
            return Path(temp_file.name)


def _replace_class_with_concept_cell(sheet: Worksheet) -> None:
    """
    Replaces the word "Class" with "Concept" in the first row of the given sheet.

    Args:
        sheet (Worksheet): The sheet in which to replace the word "Class".
    """
    for row in sheet.iter_rows():
        for cell in row:
            if cell.value == "Class":
                cell.value = "Concept"


def _replace_legacy_constraint_form(sheet: Worksheet) -> None:
    """
    Replaces the legacy form of container constraints with the new form in the given sheet.

    Args:
        sheet (Worksheet): The sheet in which to replace the old form of container constraints.
    """
    column, row = find_column_and_row_with_value(sheet, "Constraint", False)

    if not column or not row:
        return None

    # Iterate over values in the constraint column and replace old form with new form
    replaced: bool = False
    for cell_row in sheet.iter_rows(min_row=row + 1, min_col=column, max_col=column):
        cell = cell_row[0]
        if cell.value is not None:  # Skip empty cells
            # Container sheet update
            if sheet.title.lower() == "containers":
                constraints = []
                for constraint in SPLIT_ON_COMMA_PATTERN.split(str(cell.value)):
                    # latest format, do nothing
                    if "require" in constraint.lower():
                        constraints.append(constraint)
                        continue

                    try:
                        container = ContainerEntity.load(constraint, space="default")
                        container_str = container.external_id if container.space == "default" else str(container)
                        constraints.append(
                            f"requires:{container.space}_{container.external_id}(require={container_str})"
                        )
                        replaced = True
                    except ValidationError:
                        constraints.append(constraint)

                cell.value = ",".join(constraints)

            # Properties sheet update
            elif sheet.title.lower() == "properties":
                constraints = []
                for constraint in SPLIT_ON_COMMA_PATTERN.split(str(cell.value)):
                    try:
                        constraint_entity = ContainerConstraintEntity.load(constraint)

                        if constraint_entity.prefix in ["uniqueness", "requires"]:
                            constraints.append(constraint)

                        # Replace old format with new format
                        else:
                            constraints.append(f"uniqueness:{constraint}")
                            replaced = True

                    # If the constraint is not valid, we keep it as is
                    # to be caught by validation later
                    except ValidationError:
                        constraints.append(constraint)

                cell.value = ",".join(constraints)

    if replaced:
        print(
            (
                "You are using a legacy container constraints format "
                f"in the {sheet.title} sheet. Please update to the latest format."
            ),
        )



from ._client import HTTPClient
from ._data_classes import (
    ErrorDetails,
    FailedRequestItems,
    FailedRequestMessage,
    FailedResponse,
    FailedResponseItems,
    HTTPMessage,
    ItemBody,
    ItemIDBody,
    ItemMessage,
    ItemsRequest,
    ParametersRequest,
    RequestMessage,
    ResponseMessage,
    SimpleBodyRequest,
    SuccessResponse,
    SuccessResponseItems,
)

__all__ = [
    "ErrorDetails",
    "FailedRequestItems",
    "FailedRequestMessage",
    "FailedResponse",
    "FailedResponseItems",
    "HTTPClient",
    "HTTPMessage",
    "ItemBody",
    "ItemIDBody",
    "ItemMessage",
    "ItemsRequest",
    "ParametersRequest",
    "RequestMessage",
    "ResponseMessage",
    "SimpleBodyRequest",
    "SuccessResponse",
    "SuccessResponseItems",
]



"""Storage abstraction for persisting data in both local filesystem and pyodide environments."""

from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import TYPE_CHECKING, Protocol

if TYPE_CHECKING:
    try:
        import asyncio

        from pyodide.webloop import PyodideTask  # type: ignore[import-not-found]
    except ImportError:
        pass

from cognite.neat._session._usage_analytics._constants import IN_PYODIDE


@dataclass
class ReadResult(ABC):
    @property
    @abstractmethod
    def is_ready(self) -> bool: ...

    @abstractmethod
    def get_data(self) -> str: ...


@dataclass
class ManualReadResult(ReadResult):
    data: str

    @property
    def is_ready(self) -> bool:
        return True

    def get_data(self) -> str:
        return self.data


class Storage(Protocol):
    """Protocol for storage implementations."""

    def read(self, key: str) -> ReadResult:
        """Read a value from storage."""
        ...

    def write(self, key: str, value: str) -> None:
        """Write a value to storage."""
        ...

    def delete(self, key: str) -> None:
        """Delete a value from storage."""
        ...


@dataclass
class FileReadResult(ReadResult):
    data: str

    @property
    def is_ready(self) -> bool:
        return True

    def get_data(self) -> str:
        return self.data


class FileSystemStorage:
    """Storage implementation using the filesystem."""

    ENCODING = "utf-8"

    def __init__(self, base_dir: Path) -> None:
        self._base_dir = base_dir

    def _get_path(self, key: str) -> Path:
        return self._base_dir / f"{key}.bin"

    def read(self, key: str) -> ReadResult:
        path = self._get_path(key)
        if path.exists():
            return FileReadResult(path.read_text(encoding=self.ENCODING))
        return FileReadResult("")

    def write(self, key: str, value: str) -> None:
        path = self._get_path(key)
        path.write_text(value, encoding=self.ENCODING)

    def delete(self, key: str) -> None:
        path = self._get_path(key)
        path.unlink(missing_ok=True)


@dataclass
class PyodideResult(ReadResult):
    task: "PyodideTask"

    @property
    def is_ready(self) -> bool:
        return self.task.done()

    def get_data(self) -> str:
        return self.task.result()


class LocalStorageAdapter:
    """
    Storage implementation using browser IndexedDB for pyodide environments.

    This adapter provides a synchronous interface to the asynchronous IndexedDB API
    by managing an asyncio event loop.
    """

    _db_name = "neat-storage"
    _store_name = "keyval"
    _db_version = 1

    def __init__(self) -> None:
        self._loop = self._get_or_create_event_loop()

    @staticmethod
    def _get_or_create_event_loop() -> "asyncio.AbstractEventLoop":
        """Gets the current asyncio event loop or creates a new one."""
        import asyncio

        try:
            return asyncio.get_event_loop()
        except RuntimeError:
            loop = asyncio.new_event_loop()
            asyncio.set_event_loop(loop)
            return loop

    def _execute_db_operation(self, operation: str, key: str, value: str | None = None) -> str:
        """Executes a database operation by creating and calling a JS function."""

        async def db_operation() -> str:
            import js  # type: ignore[import-not-found]
            from pyodide.code import run_js  # type: ignore[import-not-found]
            from pyodide.ffi import to_js  # type: ignore[import-not-found]

            # Convert Python values to JS
            js_key = to_js(key)
            js_value = to_js(value if value is not None else "")

            # Store values in js namespace temporarily to pass them safely
            js._idb_key = js_key
            js._idb_value = js_value

            try:
                # Create and execute the function directly in JavaScript context using run_js
                js_code = f"""
                (async () => {{
                    const dbName = "{self._db_name}";
                    const dbVersion = {self._db_version};
                    const storeName = "{self._store_name}";
                    const operation = "{operation}";

                    return new Promise((resolve, reject) => {{
                        const request = indexedDB.open(dbName, dbVersion);

                        request.onerror = (event) => reject(event.target.error);

                        request.onupgradeneeded = (event) => {{
                            const db = event.target.result;
                            if (!db.objectStoreNames.contains(storeName)) {{
                                db.createObjectStore(storeName);
                            }}
                        }};

                        request.onsuccess = (event) => {{
                            const db = event.target.result;
                            const mode = (operation === "read") ? "readonly" : "readwrite";
                            try {{
                                const transaction = db.transaction([storeName], mode);
                                const store = transaction.objectStore(storeName);

                                let storeRequest;
                                if (operation === "read") {{
                                    storeRequest = store.get(_idb_key);
                                }} else if (operation === "write") {{
                                    storeRequest = store.put(_idb_value, _idb_key);
                                }} else if (operation === "delete") {{
                                    storeRequest = store.delete(_idb_key);
                                }} else {{
                                    db.close();
                                    return reject(new Error(`Unknown operation: ${{operation}}`));
                                }}

                                storeRequest.onsuccess = () => resolve(storeRequest.result ?? "");
                                storeRequest.onerror = (event) => reject(event.target.error);

                                transaction.oncomplete = () => db.close();
                                transaction.onerror = (event) => reject(event.target.error);

                            }} catch (error) {{
                                db.close();
                                reject(error);
                            }}
                        }};
                    }});
                }})()
                """

                # Use run_js which properly returns an awaitable promise
                result = await run_js(js_code)
                return str(result) if result is not None else ""
            finally:
                # Clean up js namespace
                try:
                    del js._idb_key
                    del js._idb_value
                except Exception:
                    pass

        try:
            return self._loop.run_until_complete(db_operation())
        except Exception:
            return ""

    def read(self, key: str) -> ReadResult:
        return ManualReadResult(self._execute_db_operation("read", key))

    def write(self, key: str, value: str) -> None:
        self._execute_db_operation("write", key, value)

    def delete(self, key: str) -> None:
        self._execute_db_operation("delete", key)


def get_storage(base_dir: Path | None = None) -> Storage:
    """Get the appropriate storage implementation for the current environment."""
    if IN_PYODIDE:
        return LocalStorageAdapter()
    else:
        if base_dir is None:
            import tempfile

            base_dir = Path(tempfile.gettempdir()).resolve()
        return FileSystemStorage(base_dir)



from cognite.neat._data_model.models.dms import (
    ContainerPropertyDefinition,
    ContainerReference,
    ContainerRequest,
    ViewPropertyDefinition,
    ViewRequest,
)
from cognite.neat._data_model.models.dms._view_property import (
    EdgeProperty,
    ReverseDirectRelationProperty,
    ViewCoreProperty,
    ViewCorePropertyRequest,
)

from ._differ import ItemDiffer, ObjectDiffer, field_differences
from ._differ_container import ContainerPropertyDiffer
from .data_classes import (
    AddedField,
    ChangedField,
    FieldChange,
    RemovedField,
    SeverityType,
)


class ViewDiffer(ItemDiffer[ViewRequest]):
    def __init__(
        self,
        current_container_map: dict[ContainerReference, ContainerRequest],
        new_container_map: dict[ContainerReference, ContainerRequest],
        parent_path: str | None = None,
    ):
        super().__init__(parent_path)
        self._current_container_map = current_container_map
        self._new_containe_map = new_container_map

    def diff(self, current: ViewRequest, new: ViewRequest) -> list[FieldChange]:
        changes: list[FieldChange] = []
        if current.implements != new.implements:
            # Added implements
            current_implements = set(current.implements or [])
            for new_implements in new.implements or []:
                if new_implements not in current_implements:
                    changes.append(
                        AddedField(
                            item_severity=SeverityType.BREAKING,
                            field_path="implements",
                            new_value=new_implements,
                        )
                    )

            # Removed implements
            new_implements = set(new.implements or [])
            for current_implements in current.implements or []:
                if current_implements not in new_implements:
                    changes.append(
                        RemovedField(
                            item_severity=SeverityType.BREAKING,
                            field_path="implements",
                            current_value=current_implements,
                        )
                    )

            if not changes:
                # If there are no added or removed implements, it means the order has changed
                changes.append(
                    ChangedField(
                        item_severity=SeverityType.SAFE,
                        field_path="implements",
                        current_value=str(current.implements),
                        new_value=str(new.implements),
                    )
                )

        changes.extend(self._diff_name_description(current, new))

        if current.filter != new.filter:
            changes.append(
                ChangedField(
                    field_path=self._get_path("filter"),
                    item_severity=SeverityType.WARNING,
                    new_value=str(new.filter),
                    current_value=str(current.filter),
                )
            )

        changes.extend(
            # MyPy fails to recognize that ViewPropertyDefinition and
            # the union ViewRequestProperty are the same here.
            field_differences(  # type: ignore[misc]
                "properties",
                current.properties,
                new.properties,
                add_severity=SeverityType.SAFE,
                remove_severity=SeverityType.BREAKING,
                differ=ViewPropertyDiffer(
                    self._current_container_map, self._new_containe_map, self._get_path("properties")
                ),
            )
        )

        return changes


class ViewPropertyDiffer(ObjectDiffer[ViewPropertyDefinition]):
    def __init__(
        self,
        current_container_map: dict[ContainerReference, ContainerRequest],
        new_container_map: dict[ContainerReference, ContainerRequest],
        parent_path: str | None = None,
    ):
        super().__init__(parent_path)
        self._current_container_map = current_container_map
        self._new_container_map = new_container_map

    def diff(
        self,
        current: ViewPropertyDefinition,
        new: ViewPropertyDefinition,
        identifier: str,
    ) -> list[FieldChange]:
        changes: list[FieldChange] = self._diff_name_description(current, new, identifier)
        if current.connection_type != new.connection_type:
            changes.append(
                ChangedField(
                    field_path=self._get_path(f"{identifier}.connectionType"),
                    item_severity=SeverityType.BREAKING,
                    new_value=new.connection_type,
                    current_value=current.connection_type,
                )
            )
        elif isinstance(current, ViewCoreProperty) and isinstance(new, ViewCoreProperty):
            changes.extend(self._diff_core_property(current, new, identifier))

        elif isinstance(current, EdgeProperty) and isinstance(new, EdgeProperty):
            changes.extend(self._diff_edge_property(current, new, identifier))

        elif isinstance(current, ReverseDirectRelationProperty) and isinstance(new, ReverseDirectRelationProperty):
            changes.extend(self._diff_reverse_direct_relation_property(current, new, identifier))

        return changes

    def _diff_core_property(
        self,
        current: ViewCoreProperty,
        new: ViewCoreProperty,
        identifier: str,
    ) -> list[FieldChange]:
        changes: list[FieldChange] = []

        if (current.container, current.container_property_identifier) != (
            new.container,
            new.container_property_identifier,
        ):
            item_severity = self._get_container_change_severity(
                current.container,
                current.container_property_identifier,
                new.container,
                new.container_property_identifier,
            )

            if current.container != new.container:
                changes.append(
                    ChangedField(
                        field_path=self._get_path(f"{identifier}.container"),
                        item_severity=item_severity,
                        new_value=new.container,
                        current_value=current.container,
                    )
                )
            if current.container_property_identifier != new.container_property_identifier:
                changes.append(
                    ChangedField(
                        field_path=self._get_path(f"{identifier}.containerPropertyIdentifier"),
                        item_severity=item_severity,
                        new_value=new.container_property_identifier,
                        current_value=current.container_property_identifier,
                    )
                )

        if isinstance(current, ViewCorePropertyRequest) and isinstance(new, ViewCorePropertyRequest):
            if current.source != new.source:
                changes.append(
                    ChangedField(
                        field_path=self._get_path(f"{identifier}.source"),
                        item_severity=SeverityType.BREAKING,
                        new_value=new.source,
                        current_value=current.source,
                    )
                )

        return changes

    def _diff_edge_property(
        self,
        current: EdgeProperty,
        new: EdgeProperty,
        identifier: str,
    ) -> list[FieldChange]:
        changes: list[FieldChange] = []
        if current.source != new.source:
            changes.append(
                ChangedField(
                    field_path=self._get_path(f"{identifier}.source"),
                    item_severity=SeverityType.BREAKING,
                    new_value=new.source,
                    current_value=current.source,
                )
            )
        if current.type != new.type:
            changes.append(
                ChangedField(
                    field_path=self._get_path(f"{identifier}.type"),
                    item_severity=SeverityType.WARNING,
                    new_value=new.type,
                    current_value=current.type,
                )
            )
        if current.edge_source != new.edge_source:
            changes.append(
                ChangedField(
                    field_path=self._get_path(f"{identifier}.edgeSource"),
                    item_severity=SeverityType.BREAKING,
                    new_value=new.edge_source,
                    current_value=current.edge_source,
                )
            )
        if current.direction != new.direction:
            changes.append(
                ChangedField(
                    field_path=self._get_path(f"{identifier}.direction"),
                    item_severity=SeverityType.BREAKING,
                    new_value=new.direction,
                    current_value=current.direction,
                )
            )
        return changes

    def _diff_reverse_direct_relation_property(
        self,
        current: ReverseDirectRelationProperty,
        new: ReverseDirectRelationProperty,
        identifier: str,
    ) -> list[FieldChange]:
        changes: list[FieldChange] = []
        if current.source != new.source:
            changes.append(
                ChangedField(
                    field_path=self._get_path(f"{identifier}.source"),
                    item_severity=SeverityType.BREAKING,
                    new_value=new.source,
                    current_value=current.source,
                )
            )

        if current.through != new.through:
            changes.append(
                ChangedField(
                    field_path=self._get_path(f"{identifier}.through"),
                    item_severity=SeverityType.WARNING,
                    new_value=new.through,
                    current_value=current.through,
                )
            )

        return changes

    def _get_container_change_severity(
        self,
        current_container: ContainerReference,
        current_property_identifier: str,
        new_container: ContainerReference,
        new_property_identifier: str,
    ) -> SeverityType:
        current_container_property = self._get_container_property(
            current_container, current_property_identifier, self._current_container_map
        )
        new_container_property = self._get_container_property(
            new_container, new_property_identifier, self._new_container_map
        )
        if not current_container_property or not new_container_property:
            return SeverityType.BREAKING
        differ = ContainerPropertyDiffer(parent_path=None)
        changes = differ.diff(current_container_property, new_container_property, identifier="")
        return SeverityType.max_severity([item.severity for item in changes], default=SeverityType.SAFE)

    @staticmethod
    def _get_container_property(
        container_ref: ContainerReference,
        property_identifier: str,
        containers: dict[ContainerReference, ContainerRequest],
    ) -> ContainerPropertyDefinition | None:
        container = containers.get(container_ref)
        if not container:
            return None
        property_def = container.properties.get(property_identifier)
        if not property_def:
            return None
        return property_def



"""Shortcut to data model catalog"""

from pathlib import Path

_CATALOG = Path(__file__).parent
imf_attributes = _CATALOG / "conceptual-imf-data-model.xlsx"
hello_world_pump = _CATALOG / "hello_world_pump.xlsx"
classic_model = _CATALOG / "classic_model.xlsx"



from abc import ABC, abstractmethod

from cognite.neat._data_model.models.dms._schema import RequestSchema


class Transformer(ABC):
    """Abstract base class for data model transformers."""

    @abstractmethod
    def transform(self, data_model: RequestSchema) -> RequestSchema:
        """Transform and return the modified data model."""
        raise NotImplementedError()



from ._identifiers import NameSpace

XML_SCHEMA_NAMESPACE = NameSpace("http://www.w3.org/2001/XMLSchema#")

CDF_CDM_SPACE = "cdf_cdm"
CDF_CDM_VERSION = "v1"

COGNITE_CONCEPTS_MAIN = (
    "CogniteAsset",
    "CogniteEquipment",
    "CogniteActivity",
    "CogniteTimeSeries",
    "CogniteFile",
)

COGNITE_CONCEPTS_INTERFACES = (
    "CogniteDescribable",
    "CogniteSourceable",
    "CogniteSchedulable",
    "CogniteVisualizable",
)

COGNITE_CONCEPTS_CONFIGURATIONS = (
    "CogniteSourceSystem",
    "CogniteUnit",
    "CogniteAssetClass",
    "CogniteAssetType",
    "CogniteEquipmentType",
    "CogniteFileCategory",
)
COGNITE_CONCEPTS_ANNOTATIONS = (
    "CogniteAnnotation",
    "CogniteDiagramAnnotation",
)
COGNITE_CONCEPTS_3D = (
    "CogniteCubeMap",
    "CogniteCADRevision",
    "CognitePointCloudVolume",
    "Cognite360ImageAnnotation",
    "Cognite3DObject",
    "Cognite3DRevision",
    "Cognite360Image",
    "Cognite360ImageCollection",
    "Cognite360ImageStation",
    "CognitePointCloudModel",
    "Cognite3DTransformation",
    "Cognite360ImageModel",
    "Cognite3DModel",
    "CogniteCADModel",
    "CognitePointCloudRevision",
    "CogniteCADNode",
)

COGNITE_CONCEPTS: tuple[str, ...] = (
    *COGNITE_CONCEPTS_MAIN,
    *COGNITE_CONCEPTS_INTERFACES,
    *COGNITE_CONCEPTS_CONFIGURATIONS,
    *COGNITE_CONCEPTS_ANNOTATIONS,
    *COGNITE_CONCEPTS_3D,
)

COGNITE_SPACES = (
    CDF_CDM_SPACE,
    "cdf_idm",
    "cdf_360_image_schema",
    "cdf_3d_schema",
    "cdf_apm",
    "cdf_apps_shared",
    "cdf_cdm_3d",
    "cdf_time_series_data",
    "cdf_cdm_units",
)

COGNITE_APP_SPACES = ("CommentInstanceSpace", "IndustrialCanvasInstanceSpace", "SolutionTagsInstanceSpace", "scene")

# Defaults from https://docs.cognite.com/cdf/dm/dm_reference/dm_limits_and_restrictions#list-size-limits

DEFAULT_MAX_LIST_SIZE = 1000
DEFAULT_MAX_LIST_SIZE_DIRECT_RELATIONS = 100



from typing import Any, Literal, cast
from urllib.parse import quote

from rdflib import BNode, Graph, Namespace, URIRef
from rdflib import Literal as RdfLiteral
from rdflib.plugins.sparql import prepareQuery
from rdflib.query import ResultRow

from cognite.neat._v0.core._constants import cognite_prefixes
from cognite.neat._v0.core._data_model.models.entities._constants import Unknown
from cognite.neat._v0.core._data_model.models.entities._single_value import ConceptEntity
from cognite.neat._v0.core._issues._base import IssueList
from cognite.neat._v0.core._issues.errors._general import NeatValueError
from cognite.neat._v0.core._issues.warnings._resources import (
    ResourceRedefinedWarning,
    ResourceRetrievalWarning,
)
from cognite.neat._v0.core._utils.rdf_ import remove_namespace_from_uri, uri_to_entity_components


def parse_concepts(
    graph: Graph, query: str, parameters: set, language: str, issue_list: IssueList
) -> tuple[dict, IssueList]:
    """Parse concepts from graph

    Args:
        graph: Graph containing concept definitions
        query: SPARQL query to use for parsing concepts
        parameters: Set of parameters to extract from the query results
        language: Language to use for parsing, by default "en"
        issue_list: List to collect issues during parsing

    Returns:
        Dataframe containing owl classes
    """

    concepts: dict[str, dict] = {}

    query = prepareQuery(query.format(language=language), initNs={k: v for k, v in graph.namespaces()})
    prefixes = cognite_prefixes()

    for raw in graph.query(query):
        res: dict = convert_rdflib_content(
            cast(ResultRow, raw).asdict(), uri_handling="as-concept-entity", prefixes=prefixes
        )
        res = {key: res.get(key, None) for key in parameters}

        # Safeguarding against incomplete semantic definitions
        if res["implements"] and isinstance(res["implements"], BNode):
            issue_list.append(
                ResourceRetrievalWarning(
                    res["concept"],
                    "implements",
                    error=("Unable to determine concept that is being implemented"),
                )
            )
            continue

        # sanitize the concept and implements
        res["concept"] = sanitize_entity(res["concept"])
        res["implements"] = sanitize_entity(res["implements"]) if res["implements"] else None

        concept_id = res["concept"]

        if concept_id not in concepts:
            concepts[concept_id] = res
        else:
            # Handling implements
            if concepts[concept_id]["implements"] and isinstance(concepts[concept_id]["implements"], list):
                if res["implements"] and res["implements"] not in concepts[concept_id]["implements"]:
                    concepts[concept_id]["implements"].append(res["implements"])

            elif concepts[concept_id]["implements"] and isinstance(concepts[concept_id]["implements"], str):
                concepts[concept_id]["implements"] = [concepts[concept_id]["implements"]]

                if res["implements"] and res["implements"] not in concepts[concept_id]["implements"]:
                    concepts[concept_id]["implements"].append(res["implements"])
            elif res["implements"]:
                concepts[concept_id]["implements"] = [res["implements"]]

            handle_meta("concept", concepts, concept_id, res, "name", issue_list)
            handle_meta("concept", concepts, concept_id, res, "description", issue_list)
    if not concepts:
        issue_list.append(NeatValueError("Unable to parse concepts"))

    return concepts, issue_list


def parse_properties(
    graph: Graph, query: str, parameters: set, language: str, issue_list: IssueList
) -> tuple[dict, IssueList]:
    """Parse properties from graph

    Args:
        graph: Graph containing property definitions
        query: SPARQL query to use for parsing properties
        parameters: Set of parameters to extract from the query results
        language: Language to use for parsing, by default "en"
        issue_list: List to collect issues during parsing

    Returns:
        Dataframe containing owl classes
    """

    properties: dict[str, dict] = {}

    query = prepareQuery(query.format(language=language), initNs={k: v for k, v in graph.namespaces()})
    prefixes = cognite_prefixes()

    for raw in graph.query(query):
        res: dict = convert_rdflib_content(
            cast(ResultRow, raw).asdict(), uri_handling="as-concept-entity", prefixes=prefixes
        )
        res = {key: res.get(key, None) for key in parameters}

        # Quote the concept id to ensure it is web-safe
        res["property_"] = quote(res["property_"], safe="")
        property_id = res["property_"]

        # Skip Bnode
        if isinstance(res["concept"], BNode):
            issue_list.append(
                ResourceRetrievalWarning(
                    property_id,
                    "property",
                    error="Cannot determine concept of property as it is a blank node",
                )
            )
            continue

        # Skip Bnode
        if isinstance(res["value_type"], BNode):
            issue_list.append(
                ResourceRetrievalWarning(
                    property_id,
                    "property",
                    error="Unable to determine value type of property as it is a blank node",
                )
            )
            continue

        # Quote the concept and value_type if they exist if not signal neat that they are not available
        res["concept"] = sanitize_entity(res["concept"]) if res["concept"] else str(Unknown)
        res["value_type"] = sanitize_entity(res["value_type"]) if res["value_type"] else str(Unknown)

        id_ = f"{res['concept']}.{res['property_']}"

        if id_ not in properties:
            properties[id_] = res
            properties[id_]["value_type"] = [properties[id_]["value_type"]]
        else:
            handle_meta("property", properties, id_, res, "name", issue_list)
            handle_meta(
                "property",
                properties,
                id_,
                res,
                "description",
                issue_list,
            )

            # Handling multi-value types
            if res["value_type"] not in properties[id_]["value_type"]:
                properties[id_]["value_type"].append(res["value_type"])

    for prop in properties.values():
        prop["value_type"] = ", ".join(prop["value_type"])

    if not properties:
        issue_list.append(NeatValueError("Unable to parse properties"))

    return properties, issue_list


def handle_meta(
    resource_type: str,
    resources: dict[str, dict],
    resource_id: str,
    res: dict,
    feature: str,
    issue_list: IssueList,
) -> None:
    if not resources[resource_id][feature] and res[feature]:
        resources[resource_id][feature] = res[feature]

    current_value = resources[resource_id][feature]
    new_value = res[feature]

    if not current_value and new_value:
        resources[resource_id][feature] = new_value
    elif current_value and new_value and current_value != new_value:
        issue_list.append(
            ResourceRedefinedWarning(
                identifier=resource_id,
                resource_type=resource_type,
                feature=feature,
                current_value=current_value,
                new_value=new_value,
            )
        )


def convert_rdflib_content(
    content: RdfLiteral | URIRef | dict | list,
    uri_handling: Literal["skip", "remove-namespace", "as-concept-entity"] = "skip",
    prefixes: dict[str, Namespace] | None = None,
) -> Any:
    """Converts rdflib content to a more Python-friendly format.

    Args:
        content: The content to convert, can be a RdfLiteral, URIRef, dict, or list.
        uri_handling: How to handle URIs. Options are:
            - "skip": Leave URIs as is.
            - "remove-namespace": Remove the namespace from URIs.
            - "short-form": Convert URIs to a short form using prefixes.

    """
    if isinstance(content, RdfLiteral):
        return content.toPython()
    elif isinstance(content, URIRef):
        if uri_handling == "remove-namespace":
            return remove_namespace_from_uri(content)
        elif uri_handling == "as-concept-entity":
            if components := uri_to_entity_components(content, prefixes or {}):
                return ConceptEntity(prefix=components[0], suffix=components[3], version=components[2])
            # fallback to "remove-namespace"
            else:
                return convert_rdflib_content(content, uri_handling="remove-namespace", prefixes=prefixes)
        else:
            return content.toPython()
    elif isinstance(content, dict):
        return {key: convert_rdflib_content(value, uri_handling, prefixes) for key, value in content.items()}
    elif isinstance(content, list):
        return [convert_rdflib_content(item, uri_handling, prefixes) for item in content]
    else:
        return content


def sanitize_entity(entity: str | ConceptEntity, safe: str = "") -> str:
    """Sanitize an entity to ensure it yields entity form that will pass downstream validation.

    Args:
        entity: The entity to sanitize. Can be a string or a ConceptEntity.
        safe: Characters that should not be quoted during sanitization.

    Returns:
        A web-safe string representation of the entity
    """
    if isinstance(entity, str):
        return quote(entity, safe=safe)
    # if it already we dont need to quote it so we return its string representation
    elif isinstance(entity, ConceptEntity):
        return str(entity)
    else:
        raise ValueError(f"Invalid entity type: {type(entity)}. Expected str, ConceptEntity.")



import importlib
import inspect
from abc import ABC
from types import ModuleType
from typing import TypeVar

from cognite.neat import _version
from cognite.neat._exceptions import NeatImportError


def get_current_neat_version() -> str:
    return _version.__version__


T_Cls = TypeVar("T_Cls")


def get_concrete_subclasses(base_cls: type[T_Cls], exclude_direct_abc_inheritance: bool = True) -> list[type[T_Cls]]:
    """
    Returns a list of all concrete subclasses of the given base class.
    Args:
        base_cls (type[T_Cls]): The base class to find subclasses for.
        exclude_direct_abc_inheritance (bool): If True, excludes classes that directly inherit from `abc.ABC`.
            This is used as a marker to filter out intermediate base classes. Defaults to True.
    Returns:
        list[type[T_Cls]]: A list of concrete subclasses of the base class.
    """
    to_check = [base_cls]
    subclasses: list[type[T_Cls]] = []
    seen: set[type[T_Cls]] = {base_cls}
    while to_check:
        current_cls = to_check.pop()
        for subclass in current_cls.__subclasses__():
            if subclass in seen:
                continue
            if (not inspect.isabstract(subclass)) and (
                not exclude_direct_abc_inheritance or ABC not in subclass.__bases__
            ):
                subclasses.append(subclass)
            seen.add(subclass)
            to_check.append(subclass)
    return subclasses


def local_import(module: str, extra: str) -> ModuleType:
    try:
        return importlib.import_module(module)
    except ImportError as e:
        raise NeatImportError(module.split(".")[0], extra) from e



from collections.abc import Callable

from cognite.client.credentials import CredentialProvider, OAuthClientCredentials, OAuthInteractive, Token

from cognite.neat._utils.text import humanize_collection

from .env_vars import ClientEnvironmentVariables, LoginFlow


def get_credentials(env_vars: ClientEnvironmentVariables) -> CredentialProvider:
    options: dict[LoginFlow, Callable[[ClientEnvironmentVariables], CredentialProvider]] = {
        "client_credentials": create_client_credentials,
        "interactive": create_interactive_credentials,
        "token": create_token_credentials,
    }
    return options[env_vars.LOGIN_FLOW](env_vars)


def create_client_credentials(env_vars: ClientEnvironmentVariables) -> CredentialProvider:
    missing: list[str] = []
    if not env_vars.IDP_CLIENT_ID:
        missing.append("IDP_CLIENT_ID")
    if not env_vars.IDP_CLIENT_SECRET:
        missing.append("IDP_CLIENT_SECRET")
    if env_vars.IDP_CLIENT_ID is None or env_vars.IDP_CLIENT_SECRET is None:
        raise ValueError(
            f"The following environment variables must be set for "
            f"client credentials authentication: {humanize_collection(missing)}"
        )

    if env_vars.PROVIDER == "cdf":
        return OAuthClientCredentials(
            client_id=env_vars.IDP_CLIENT_ID,
            client_secret=env_vars.IDP_CLIENT_SECRET,
            token_url=env_vars.idp_token_url,
            scopes=None,  # type: ignore[arg-type]
        )
    return OAuthClientCredentials(
        client_id=env_vars.IDP_CLIENT_ID,
        client_secret=env_vars.IDP_CLIENT_SECRET,
        token_url=env_vars.idp_token_url,
        audience=env_vars.idp_audience,
        scopes=env_vars.idp_scopes,
    )


def create_interactive_credentials(env_vars: ClientEnvironmentVariables) -> CredentialProvider:
    if not env_vars.IDP_CLIENT_ID:
        raise ValueError("IDP_CLIENT_ID environment variable must be set for interactive authentication.")
    return OAuthInteractive(
        client_id=env_vars.IDP_CLIENT_ID,
        authority_url=env_vars.idp_authority_url,
        scopes=env_vars.idp_scopes,
    )


def create_token_credentials(env_vars: ClientEnvironmentVariables) -> CredentialProvider:
    if not env_vars.CDF_TOKEN:
        raise ValueError("CDF_TOKEN environment variable must be set for token authentication.")
    return Token(env_vars.CDF_TOKEN)



import json
from collections import defaultdict
from dataclasses import dataclass, field
from typing import Any, Literal

from pydantic import ValidationError

from cognite.neat._data_model._constants import DEFAULT_MAX_LIST_SIZE, DEFAULT_MAX_LIST_SIZE_DIRECT_RELATIONS
from cognite.neat._data_model.importers._table_importer.data_classes import (
    CREATOR_KEY,
    CREATOR_MARKER,
    DMSContainer,
    DMSEnum,
    DMSNode,
    DMSProperty,
    DMSView,
    EntityTableFilter,
    MetadataValue,
    RAWFilterTableFilter,
    TableDMS,
    TableViewFilter,
)
from cognite.neat._data_model.models.dms import (
    ContainerPropertyDefinition,
    ContainerReference,
    ContainerRequest,
    DataModelRequest,
    DataType,
    DirectNodeRelation,
    EnumProperty,
    EqualsFilterData,
    Filter,
    FilterAdapter,
    HasDataFilter,
    InFilterData,
    ListablePropertyTypeDefinition,
    NodeReference,
    RequestSchema,
    RequiresConstraintDefinition,
    UniquenessConstraintDefinition,
    ViewCorePropertyRequest,
    ViewReference,
    ViewRequest,
    ViewRequestProperty,
)
from cognite.neat._data_model.models.dms._view_property import (
    EdgeProperty,
    MultiEdgeProperty,
    MultiReverseDirectRelationPropertyRequest,
    ReverseDirectRelationProperty,
    SingleEdgeProperty,
    SingleReverseDirectRelationPropertyRequest,
)
from cognite.neat._data_model.models.entities import ParsedEntity


@dataclass
class ViewProperties:
    properties: list[DMSProperty] = field(default_factory=list)
    nodes: list[DMSNode] = field(default_factory=list)


@dataclass
class ContainerProperties:
    properties_by_id: dict[tuple[ContainerReference, str], dict] = field(default_factory=dict)
    enum_collections: list[DMSEnum] = field(default_factory=list)


class DMSTableWriter:
    def __init__(self, default_space: str, default_version: str, skip_properties_in_other_spaces: bool) -> None:
        self.default_space = default_space
        self.default_version = default_version
        self.skip_properties_in_other_spaces = skip_properties_in_other_spaces

    ## Main Entry Point ###
    def write_tables(self, schema: RequestSchema) -> TableDMS:
        metadata = self.write_metadata(schema.data_model)
        container_properties = self.write_container_properties(schema.containers)
        view_properties = self.write_view_properties(schema.views, container_properties)
        views = self.write_views(schema.views)
        containers = self.write_containers(schema.containers)

        return TableDMS(
            metadata=metadata,
            properties=view_properties.properties,
            views=views,
            containers=containers,
            enum=container_properties.enum_collections,
            nodes=view_properties.nodes,
        )

    ### Metadata Sheet ###
    @classmethod
    def write_metadata(cls, data_model: DataModelRequest) -> list[MetadataValue]:
        metadata = [
            MetadataValue(key=key, value=value)
            for key, value in data_model.model_dump(
                mode="json", by_alias=True, exclude_none=True, exclude={"views", "description"}
            ).items()
        ]
        if data_model.description:
            description, creator = cls._serialize_description(data_model.description)
            if description:
                metadata.append(MetadataValue(key="description", value=description))
            if creator:
                metadata.append(MetadataValue(key=CREATOR_KEY, value=creator))
        return metadata

    @staticmethod
    def _serialize_description(description: str | None) -> tuple[str | None, str | None]:
        """DataModelRequest does not have a 'creator' field, this is a special addition that the Neat tables
        format supports (and recommends using). If the data model was created using Neat, the suffix of the
        description will be Creator: <creator>. This function extracts that information."""
        if description is None:
            return None, None
        if CREATOR_MARKER not in description:
            return description, None

        description, creator = description.rsplit(CREATOR_MARKER, 1)
        return description.rstrip(), creator.strip()

    ### Container Properties Sheet ###

    def write_containers(self, containers: list[ContainerRequest]) -> list[DMSContainer]:
        return [
            DMSContainer(
                container=self._create_container_entity(container),
                name=container.name,
                description=container.description,
                constraint=self._create_container_constraints(container),
                used_for=container.used_for,
            )
            for container in containers
        ]

    def write_container_properties(self, containers: list[ContainerRequest]) -> ContainerProperties:
        indices_by_container_property = self._write_container_indices(containers)
        constraints_by_container_property = self._write_container_property_constraints(containers)

        output = ContainerProperties()
        for container in containers:
            for prop_id, prop in container.properties.items():
                container_property = self._write_container_property(
                    container.as_reference(),
                    prop_id,
                    prop,
                    indices_by_container_property,
                    constraints_by_container_property,
                )
                output.properties_by_id[(container.as_reference(), prop_id)] = container_property
                if isinstance(prop.type, EnumProperty):
                    output.enum_collections.extend(
                        self._write_enum_collection(container.as_reference(), prop_id, prop.type)
                    )
        return output

    def _write_container_property(
        self,
        container_ref: ContainerReference,
        prop_id: str,
        prop: ContainerPropertyDefinition,
        indices_by_container_property: dict[tuple[ContainerReference, str], list[ParsedEntity]],
        constraints_by_container_property: dict[tuple[ContainerReference, str], list[ParsedEntity]],
    ) -> dict[str, Any]:
        return dict(
            connection=self._write_container_property_connection(prop.type),
            value_type=self._write_container_property_value_type(prop, prop_id, container_ref),
            min_count=0 if prop.nullable else 1,
            max_count=self._write_container_property_max_count(prop.type),
            immutable=prop.immutable,
            default=json.dumps(prop.default_value) if isinstance(prop.default_value, dict) else prop.default_value,
            auto_increment=prop.auto_increment,
            container=self._create_container_entity(container_ref),
            container_property=prop_id,
            container_property_name=prop.name,
            container_property_description=prop.description,
            index=indices_by_container_property.get((container_ref, prop_id)),
            constraint=constraints_by_container_property.get((container_ref, prop_id)),
        )

    def _write_container_property_connection(self, dtype: DataType) -> ParsedEntity | None:
        if not isinstance(dtype, DirectNodeRelation):
            return None
        properties: dict[str, str] = {}
        if dtype.container is not None:
            properties["container"] = str(self._create_container_entity(dtype.container))
        return ParsedEntity("", "direct", properties=properties)

    def _write_container_property_value_type(
        self, prop: ContainerPropertyDefinition, prop_id: str, container_ref: ContainerReference
    ) -> ParsedEntity:
        if isinstance(prop.type, DirectNodeRelation):
            # Will be overwritten if the view property has source set.
            return ParsedEntity("", "#N/A", properties={})
        elif isinstance(prop.type, EnumProperty):
            enum_properties = {"collection": self._enum_collection_name(container_ref, prop_id)}
            if prop.type.unknown_value is not None:
                enum_properties["unknownValue"] = prop.type.unknown_value
            return ParsedEntity("", "enum", properties=enum_properties)
        elif isinstance(prop.type, ListablePropertyTypeDefinition):
            # List and maxListSize are included in the maxCount of the property, so we exclude them here.
            entity_properties = prop.type.model_dump(
                mode="json", by_alias=True, exclude={"list", "maxListSize", "type"}, exclude_none=True
            )
            return ParsedEntity("", prop.type.type, properties=entity_properties)
        else:
            # Should not happen as all types are either ListablePropertyTypeDefinition or EnumProperty.
            return ParsedEntity("", prop.type.type, properties={})

    @staticmethod
    def _write_container_property_max_count(dtype: DataType) -> int | None:
        if isinstance(dtype, ListablePropertyTypeDefinition) and dtype.list:
            if dtype.max_list_size is not None:
                return dtype.max_list_size
            elif isinstance(dtype, DirectNodeRelation):
                return DEFAULT_MAX_LIST_SIZE_DIRECT_RELATIONS
            else:
                return DEFAULT_MAX_LIST_SIZE
        return 1

    @staticmethod
    def _write_container_indices(
        containers: list[ContainerRequest],
    ) -> dict[tuple[ContainerReference, str], list[ParsedEntity]]:
        """Writes container indices and groups them by (container_reference, property_id)."""
        indices_by_id: dict[tuple[ContainerReference, str], list[ParsedEntity]] = defaultdict(list)
        for container in containers:
            if not container.indexes:
                continue
            for index_id, index in container.indexes.items():
                for order, prop_id in enumerate(index.properties, 1):
                    entity_properties = index.model_dump(
                        mode="json", by_alias=True, exclude={"index_type", "properties"}, exclude_none=True
                    )
                    if len(index.properties) > 1:
                        entity_properties["order"] = str(order)
                    entity = ParsedEntity(index.index_type, index_id, properties=entity_properties)
                    indices_by_id[(container.as_reference(), prop_id)].append(entity)
        return indices_by_id

    @staticmethod
    def _write_container_property_constraints(
        containers: list[ContainerRequest],
    ) -> dict[tuple[ContainerReference, str], list[ParsedEntity]]:
        """Writes container constraints and groups them by (container_reference, property_id).

        Note this only includes uniqueness constraints, the require constraints is handled
        in the writing of the container itself.
        """
        constraints_by_id: dict[tuple[ContainerReference, str], list[ParsedEntity]] = defaultdict(list)
        for container in containers:
            if not container.constraints:
                continue
            for constraint_id, constraint in container.constraints.items():
                if not isinstance(constraint, UniquenessConstraintDefinition):
                    continue
                for order, prop_id in enumerate(constraint.properties, 1):
                    entity_properties = constraint.model_dump(
                        mode="json", by_alias=True, exclude={"constraint_type", "properties"}, exclude_none=True
                    )
                    if len(constraint.properties) > 1:
                        entity_properties["order"] = str(order)
                    entity = ParsedEntity(constraint.constraint_type, constraint_id, properties=entity_properties)
                    constraints_by_id[(container.as_reference(), prop_id)].append(entity)
        return constraints_by_id

    def _create_container_constraints(self, container: ContainerRequest) -> list[ParsedEntity] | None:
        if not container.constraints:
            return None
        output: list[ParsedEntity] = []
        for constraint_id, constraint in container.constraints.items():
            if not isinstance(constraint, RequiresConstraintDefinition):
                continue
            entity_properties = {"require": str(self._create_container_entity(constraint.require))}
            output.append(
                ParsedEntity(prefix=constraint.constraint_type, suffix=constraint_id, properties=entity_properties)
            )
        return output or None

    ### Enum Sheet ###
    @staticmethod
    def _enum_collection_name(container_ref: ContainerReference, prop_id: str) -> str:
        return f"{container_ref.external_id}.{prop_id}"

    def _write_enum_collection(
        self, container_ref: ContainerReference, prop_id: str, enum: EnumProperty
    ) -> list[DMSEnum]:
        output: list[DMSEnum] = []
        name = self._enum_collection_name(container_ref, prop_id)
        for value_id, value in enum.values.items():
            output.append(
                DMSEnum(
                    collection=name,
                    value=value_id,
                    name=value.name,
                    description=value.description,
                )
            )
        return output

    ### View Sheet ###
    def write_views(self, views: list[ViewRequest]) -> list[DMSView]:
        return [
            DMSView(
                view=self._create_view_entity(view),
                name=view.name,
                description=view.description,
                implements=[self._create_view_entity(parent) for parent in view.implements]
                if view.implements
                else None,
                filter=self.write_view_filter(view.filter),
            )
            for view in views
        ]

    def write_view_filter(self, filter: Filter | None) -> TableViewFilter | None:
        if filter is None:
            return None
        filter_type, entities = self._get_entity_filter(filter)
        if filter_type is not None and entities:
            return EntityTableFilter(type=filter_type, entities=entities)
        else:
            return RAWFilterTableFilter(filter=FilterAdapter.dump_json(filter, by_alias=True).decode(encoding="utf-8"))

    def _get_entity_filter(self, filter: Filter) -> tuple[Literal["nodeType", "hasData"] | None, list[ParsedEntity]]:
        """If the filter is an entity-based filter (Equals or In on nodes), return the type and entities.
        Otherwise, return (None, [])."""
        if filter is None or len(filter) != 1:
            return None, []
        filter_name, body = next(iter(filter.items()))
        if (
            isinstance(body, EqualsFilterData)
            and body.property == ["node", "type"]
            and isinstance(body.value, dict)
            and (node_reference := self._try_get_node_reference(body.value))
        ):
            return "nodeType", [self._create_node_entity(node_reference)]
        elif (
            isinstance(body, InFilterData)
            and body.property == ["node", "type"]
            and isinstance(body.values, list)
            # All values must be node references
            and len(node_references := [ref for value in body.values if (ref := self._try_get_node_reference(value))])
            == len(body.values)
        ):
            return "nodeType", [self._create_node_entity(node) for node in node_references]
        elif (
            isinstance(body, HasDataFilter)
            and
            # All data must be container references, a single view reference makes it a raw filter
            len(container_refs := [item for item in body.data if isinstance(item, ContainerReference)])
            == len(body.data)
        ):
            return "hasData", [self._create_container_entity(item) for item in container_refs]
        else:
            return None, []

    @staticmethod
    def _try_get_node_reference(value: Any) -> NodeReference | None:
        try:
            return NodeReference.model_validate(value)
        except ValidationError:
            return None

    def write_view_properties(self, views: list[ViewRequest], container: ContainerProperties) -> ViewProperties:
        output = ViewProperties()
        for view in views:
            if not view.properties:
                continue
            if self.skip_properties_in_other_spaces and view.space != self.default_space:
                continue
            for prop_id, prop in view.properties.items():
                output.properties.append(self._write_view_property(view, prop_id, prop, container))
                if isinstance(prop, EdgeProperty):
                    output.nodes.append(self._write_node(prop))
        return output

    def _write_view_property(
        self, view: ViewRequest, prop_id: str, prop: ViewRequestProperty, container: ContainerProperties
    ) -> DMSProperty:
        container_properties: dict[str, Any] = {}
        if isinstance(prop, ViewCorePropertyRequest):
            identifier = (prop.container, prop.container_property_identifier)
            if identifier in container.properties_by_id:
                container_properties = container.properties_by_id[identifier]
        view_properties: dict[str, Any] = dict(
            view=self._create_view_entity(view), view_property=prop_id, name=prop.name, description=prop.description
        )
        if connection := self._write_view_property_connection(prop):
            view_properties["connection"] = connection
        if view_value_type := self._write_view_property_value_type(prop):
            view_properties["value_type"] = view_value_type
        view_min_count = self._write_view_property_min_count(prop)
        if view_min_count is not None:
            view_properties["min_count"] = view_min_count
        view_max_count = self._write_view_property_max_count(prop)
        if view_max_count != "container":
            view_properties["max_count"] = view_max_count

        # Overwrite container properties with view properties where relevant.
        args = container_properties | view_properties
        return DMSProperty(**args)

    def _write_view_property_connection(self, prop: ViewRequestProperty) -> ParsedEntity | None:
        if isinstance(prop, ViewCorePropertyRequest):
            # Use the container definition for connection
            return None
        elif isinstance(prop, EdgeProperty):
            edge_properties: dict[str, str] = {}
            if prop.direction != "outwards":
                edge_properties["direction"] = prop.direction
            if prop.edge_source is not None:
                edge_properties["edgeSource"] = str(self._create_view_entity(prop.edge_source))
            edge_properties["type"] = str(self._create_node_entity(prop.type))
            return ParsedEntity("", "edge", properties=edge_properties)
        elif isinstance(prop, ReverseDirectRelationProperty):
            return ParsedEntity("", "reverse", properties={"property": prop.through.identifier})
        else:
            raise ValueError(f"Unknown view property type: {type(prop)}")

    def _write_view_property_value_type(self, prop: ViewRequestProperty) -> ParsedEntity | None:
        if isinstance(prop, ViewCorePropertyRequest):
            if prop.source:
                return self._create_view_entity(prop.source)
            else:
                # Use the container definition for value type
                return None
        elif isinstance(prop, ReverseDirectRelationProperty | EdgeProperty):
            return self._create_view_entity(prop.source)
        else:
            raise ValueError(f"Unknown view property type: {type(prop)}")

    @staticmethod
    def _write_view_property_min_count(prop: ViewRequestProperty) -> int | None:
        if isinstance(prop, ViewCorePropertyRequest):
            # Use the container definition for min count
            return None
        # Edges and reverse relations cannot be required.
        return 0

    @staticmethod
    def _write_view_property_max_count(prop: ViewRequestProperty) -> int | None | Literal["container"]:
        if isinstance(prop, ViewCorePropertyRequest):
            # Use the container definition for max count
            return "container"
        elif isinstance(prop, SingleEdgeProperty | SingleReverseDirectRelationPropertyRequest):
            return 1
        elif isinstance(prop, MultiEdgeProperty | MultiReverseDirectRelationPropertyRequest):
            return None
        else:
            raise ValueError(f"Unknown view property type: {type(prop)}")

    ### Node Sheet ###

    def _write_node(self, prop: EdgeProperty) -> DMSNode:
        return DMSNode(node=self._create_node_entity(prop.type))

    ## Entity Helpers ###

    def _create_view_entity(self, view: ViewRequest | ViewReference) -> ParsedEntity:
        prefix = view.space
        properties = {"version": view.version}
        if view.space == self.default_space:
            prefix = ""
            if view.version == self.default_version:
                # Only use default version if space is also default.
                properties = {}
        return ParsedEntity(prefix=prefix, suffix=view.external_id, properties=properties)

    def _create_container_entity(self, container: ContainerRequest | ContainerReference) -> ParsedEntity:
        prefix = container.space
        if container.space == self.default_space:
            prefix = ""
        return ParsedEntity(prefix=prefix, suffix=container.external_id, properties={})

    def _create_node_entity(self, node: NodeReference) -> ParsedEntity:
        prefix = node.space
        if node.space == self.default_space:
            prefix = ""
        return ParsedEntity(prefix=prefix, suffix=node.external_id, properties={})



import uuid

from cognite.neat._client import NeatClient
from cognite.neat._config import NeatConfig
from cognite.neat._data_model.rules.cdf._orchestrator import CDFRulesOrchestrator
from cognite.neat._session._wrappers import session_wrapper
from cognite.neat._store._store import NeatStore

from ._html._render import render


@session_wrapper
class CDF:
    """Read from a data source into NeatSession graph store."""

    def __init__(self, store: NeatStore, client: NeatClient, config: NeatConfig) -> None:
        self._store = store
        self._client = client
        self._config = config

    def _repr_html_(self) -> str:
        """Generate HTML representation of CDF schema statistics."""
        unique_id = str(uuid.uuid4())[:8]

        return render(
            "statistics",
            {
                "unique_id": unique_id,
                "spaces_current": self._store.cdf_limits.spaces.count,
                "spaces_limit": self._store.cdf_limits.spaces.limit,
                "containers_current": self._store.cdf_limits.containers.count,
                "containers_limit": self._store.cdf_limits.containers.limit,
                "views_current": self._store.cdf_limits.views.count,
                "views_limit": self._store.cdf_limits.views.limit,
                "data_models_current": self._store.cdf_limits.data_models.count,
                "data_models_limit": self._store.cdf_limits.data_models.limit,
            },
        )

    def analyze(self) -> None:
        """Analyze the entity of CDF data models."""

        on_success = CDFRulesOrchestrator(
            limits=self._store.cdf_limits,
            space_statistics=self._store.cdf_space_statistics,
            can_run_validator=self._config.validation.can_run_validator,
            enable_alpha_validators=self._config.alpha.enable_experimental_validators,
        )

        return self._store.cdf_analyze(on_success)



import itertools
from collections import Counter, defaultdict

from cognite.neat._v0.core._constants import get_base_concepts
from cognite.neat._v0.core._data_model._constants import PATTERNS, EntityTypes
from cognite.neat._v0.core._data_model.models._import_contexts import ImportContext, SpreadsheetContext
from cognite.neat._v0.core._data_model.models.entities import ConceptEntity, UnknownEntity
from cognite.neat._v0.core._data_model.models.entities._multi_value import MultiValueTypeInfo
from cognite.neat._v0.core._issues import IssueList
from cognite.neat._v0.core._issues.errors import NeatValueError
from cognite.neat._v0.core._issues.errors._resources import (
    ResourceDuplicatedError,
    ResourceNotDefinedError,
)
from cognite.neat._v0.core._issues.warnings._models import (
    ConceptOnlyDataModelWarning,
    ConversionToPhysicalModelImpossibleWarning,
    DanglingPropertyWarning,
    UndefinedConceptWarning,
)
from cognite.neat._v0.core._issues.warnings._resources import (
    ResourceNotDefinedWarning,
    ResourceRegexViolationWarning,
)
from cognite.neat._v0.core._utils.text import humanize_collection

from ._verified import ConceptualDataModel


class ConceptualValidation:
    """This class does all the validation of the conceptual data model that have dependencies
    between components."""

    def __init__(
        self,
        data_model: ConceptualDataModel,
        context: ImportContext | None = None,
    ):
        # import here to avoid circular import issues
        from cognite.neat._v0.core._data_model.analysis._base import DataModelAnalysis

        self.data_model = data_model
        self.analysis = DataModelAnalysis(self.data_model)
        self._read_info_by_spreadsheet = context if isinstance(context, SpreadsheetContext) else SpreadsheetContext({})
        self._metadata = data_model.metadata
        self._properties = data_model.properties
        self._concepts = data_model.concepts
        self._cdf_concepts = {
            ConceptEntity.load(concept_as_string) for concept_as_string in get_base_concepts(base_model="CogniteCore")
        }
        self.issue_list = IssueList()

    def validate(self) -> IssueList:
        self._duplicated_resources()
        self._namespaces_reassigned()
        self._concepts_without_properties_exist()
        self._concepts_with_properties_defined()
        self._ancestors_defined()

        self._object_properties_use_defined_concepts()
        self._dangling_properties()

        self._concept_only_data_model()
        self._regex_compliance_with_physical_data_model()
        self._physical_data_model_conversion()

        return self.issue_list

    def _physical_data_model_conversion(self) -> None:
        """Check if the conceptual data model has issues that will likely lead
        to problems when converting to a physical data model."""
        warning_types_preventing_conversion = [
            ConceptOnlyDataModelWarning,
            ResourceRegexViolationWarning,
            ResourceNotDefinedWarning,
            UndefinedConceptWarning,
            DanglingPropertyWarning,
        ]

        if seen_warnings := frozenset(
            [
                warning_type.__name__
                for warning_type in warning_types_preventing_conversion
                if self.issue_list.has_warning_type(warning_type)
            ]
        ):
            self.issue_list.append_if_not_exist(ConversionToPhysicalModelImpossibleWarning(issue_types=seen_warnings))

    def _concept_only_data_model(self) -> None:
        """Check if the data model only consists of concepts without any properties."""
        if not self._properties:
            self.issue_list.append_if_not_exist(ConceptOnlyDataModelWarning())

    def _dangling_properties(self) -> None:
        """Check if there are properties that do not reference any concept."""
        dangling_properties = [prop for prop in self._properties if prop.concept == UnknownEntity()]
        if dangling_properties:
            for prop in dangling_properties:
                self.issue_list.append_if_not_exist(DanglingPropertyWarning(property_id=prop.property_))

    def _duplicated_resources(self) -> None:
        properties_sheet_info = self._read_info_by_spreadsheet.get("Properties")
        concepts_sheet_info = self._read_info_by_spreadsheet.get("Concepts")

        visited = defaultdict(list)
        for row_no, property_ in enumerate(self._properties):
            visited[property_._identifier()].append(
                properties_sheet_info.adjusted_row_number(row_no) if properties_sheet_info else row_no + 1
            )

        for identifier, rows in visited.items():
            if len(rows) == 1:
                continue
            self.issue_list.append(
                ResourceDuplicatedError(
                    identifier[1],
                    "property",
                    (
                        "the Properties sheet at row "
                        f"{humanize_collection(rows)}"
                        " if data model is read from a spreadsheet."
                    ),
                )
            )

        visited = defaultdict(list)
        for row_no, concept in enumerate(self._concepts):
            visited[concept._identifier()].append(
                concepts_sheet_info.adjusted_row_number(row_no) if concepts_sheet_info else row_no + 1
            )

        for identifier, rows in visited.items():
            if len(rows) == 1:
                continue
            self.issue_list.append_if_not_exist(
                ResourceDuplicatedError(
                    identifier[0],
                    "concept",
                    (
                        f"the Concepts sheet at row {humanize_collection(rows)}"
                        " if data model is read from a spreadsheet."
                    ),
                )
            )

    def _concepts_without_properties_exist(self) -> None:
        """This validation checks if concepts have properties defined or inherit properties from other concepts."""
        concepts = {concept.concept for concept in self._concepts}
        ancestors_by_concept = self.analysis.parents_by_concept(include_ancestors=True, include_different_space=True)
        concepts_with_properties = self.analysis.defined_concepts().union(self._cdf_concepts)

        if candidate_concepts := concepts.difference(concepts_with_properties):
            for concept in candidate_concepts:
                # Here we check if at least one of the ancestors of the concept has properties
                if (ancestors := ancestors_by_concept.get(concept)) and ancestors.intersection(
                    concepts_with_properties
                ):
                    continue

                self.issue_list.append_if_not_exist(UndefinedConceptWarning(concept_id=str(concept)))

    def _concepts_with_properties_defined(self) -> None:
        """This validation checks if concepts to which properties are attached are defined."""
        concepts = {concept.concept for concept in self._concepts}
        concepts_with_properties = {property_.concept for property_ in self._properties} - {UnknownEntity()}

        if undefined_concepts_with_properties := concepts_with_properties.difference(concepts):
            for concept in undefined_concepts_with_properties:
                self.issue_list.append_if_not_exist(
                    ResourceNotDefinedError(
                        identifier=concept,
                        resource_type="concept",
                        location="Concepts sheet",
                    )
                )

    def _ancestors_defined(self) -> None:
        """This is a validation to check if the ancestor concepts (e.g. parents) are defined."""
        concepts = {concept.concept for concept in self._concepts}.union(self._cdf_concepts)
        ancestors = set(
            itertools.chain.from_iterable(
                self.analysis.parents_by_concept(include_ancestors=True, include_different_space=True).values()
            )
        ).difference(self._cdf_concepts)

        if undefined_ancestor := ancestors.difference(concepts):
            for ancestor in undefined_ancestor:
                self.issue_list.append_if_not_exist(
                    ResourceNotDefinedWarning(
                        resource_type="concept",
                        identifier=ancestor,
                        location="Concepts sheet",
                    )
                )

    def _object_properties_use_defined_concepts(self) -> None:
        """Check if the value types of object properties are defined as concepts."""

        concepts = {concept.concept for concept in self._concepts}
        # We remove UnknownEntity from the concepts to avoid false positives
        # as `UnknownEntity` is used as a placeholder when the value type is not defined.
        value_types = {
            property_.value_type
            for property_ in self.data_model.properties
            if property_.type_ == EntityTypes.object_property
        }.difference({UnknownEntity()})

        if undefined_value_types := value_types.difference(concepts):
            for value_type in undefined_value_types:
                self.issue_list.append_if_not_exist(
                    ResourceNotDefinedWarning(
                        resource_type="concept",
                        identifier=value_type,
                        location="Concepts sheet",
                    )
                )

    def _regex_compliance_with_physical_data_model(self) -> None:
        """Check regex compliance with DMS of properties, classes and value types."""

        for prop_ in self._properties:
            if not PATTERNS.physical_property_id_compliance.match(prop_.property_):
                self.issue_list.append_if_not_exist(
                    ResourceRegexViolationWarning(
                        prop_.property_,
                        "Property",
                        "Properties sheet, Property column",
                        PATTERNS.physical_property_id_compliance.pattern,
                    )
                )
            if prop_.concept != UnknownEntity() and not PATTERNS.view_id_compliance.match(prop_.concept.suffix):
                self.issue_list.append_if_not_exist(
                    ResourceRegexViolationWarning(
                        prop_.concept,
                        "Concept",
                        "Properties sheet, Concept column",
                        PATTERNS.view_id_compliance.pattern,
                    )
                )

            # Handling Value Type
            if (
                isinstance(prop_.value_type, ConceptEntity)
                and prop_.value_type != UnknownEntity()
                and not PATTERNS.view_id_compliance.match(prop_.value_type.suffix)
            ):
                self.issue_list.append_if_not_exist(
                    ResourceRegexViolationWarning(
                        prop_.value_type,
                        "Value Type",
                        "Properties sheet, Value Type column",
                        PATTERNS.view_id_compliance.pattern,
                    )
                )
            if isinstance(prop_.value_type, MultiValueTypeInfo):
                for value_type in prop_.value_type.types:
                    if (
                        isinstance(prop_.value_type, ConceptEntity)
                        and prop_.value_type != UnknownEntity()
                        and not PATTERNS.view_id_compliance.match(value_type.suffix)
                    ):
                        self.issue_list.append_if_not_exist(
                            ResourceRegexViolationWarning(
                                value_type,
                                "Value Type",
                                "Properties sheet, Value Type column",
                                PATTERNS.view_id_compliance.pattern,
                            )
                        )

        for concepts in self._concepts:
            if not PATTERNS.view_id_compliance.match(concepts.concept.suffix):
                self.issue_list.append_if_not_exist(
                    ResourceRegexViolationWarning(
                        concepts.concept,
                        "Concept",
                        "Concepts sheet, Concept column",
                        PATTERNS.view_id_compliance.pattern,
                    )
                )

            if concepts.implements:
                for parent in concepts.implements:
                    if not PATTERNS.view_id_compliance.match(parent.suffix):
                        self.issue_list.append_if_not_exist(
                            ResourceRegexViolationWarning(
                                parent,
                                "Concept",
                                "Concepts sheet, Implements column",
                                PATTERNS.view_id_compliance.pattern,
                            )
                        )

    def _namespaces_reassigned(self) -> None:
        prefixes = self.data_model.prefixes.copy()
        prefixes[self.data_model.metadata.namespace.prefix] = self.data_model.metadata.namespace

        if len(set(prefixes.values())) != len(prefixes):
            reused_namespaces = [value for value, count in Counter(prefixes.values()).items() if count > 1]
            impacted_prefixes = [key for key, value in prefixes.items() if value in reused_namespaces]
            self.issue_list.append_if_not_exist(
                NeatValueError(
                    "Namespace collision detected. The following prefixes "
                    f"are assigned to the same namespace: {impacted_prefixes}"
                    f"\nImpacted namespaces: {reused_namespaces}"
                    "\nMake sure that each unique namespace is assigned to a unique prefix"
                )
            )



from __future__ import annotations

import itertools
from datetime import datetime
from pathlib import Path
from types import GenericAlias
from typing import Any, ClassVar, Literal, cast, get_args, overload

from openpyxl import Workbook
from openpyxl.cell import MergedCell
from openpyxl.styles import Alignment, Border, Font, PatternFill, Side
from openpyxl.utils import get_column_letter
from openpyxl.worksheet.datavalidation import DataValidation
from openpyxl.worksheet.worksheet import Worksheet
from rdflib import Namespace

from cognite.neat._v0.core._constants import BASE_MODEL, get_base_concepts
from cognite.neat._v0.core._data_model._constants import get_internal_properties
from cognite.neat._v0.core._data_model._shared import VerifiedDataModel
from cognite.neat._v0.core._data_model.models import (
    SheetRow,
)
from cognite.neat._v0.core._data_model.models._base_verified import (
    BaseVerifiedMetadata,
    RoleTypes,
)
from cognite.neat._v0.core._data_model.models.conceptual._verified import (
    ConceptualDataModel,
)
from cognite.neat._v0.core._data_model.models.data_types import (
    _DATA_TYPE_BY_DMS_TYPE,
)
from cognite.neat._v0.core._data_model.models.physical._verified import PhysicalDataModel
from cognite.neat._v0.core._utils.spreadsheet import (
    find_column_and_row_with_value,
    generate_data_validation,
)

from ._base import BaseExporter

MAX_COLUMN_WIDTH = 70.0


class ExcelExporter(BaseExporter[VerifiedDataModel, Workbook]):
    """Export data_model to Excel.

    Args:
        styling: The styling to use for the Excel file. Defaults to "default". See below for details
            on the different styles.
        new_model_id: The new model ID to use for the exported spreadsheet. This is only applicable if the input
            data_model have 'is_reference' set. If provided, the model ID will be used to automatically create the
            new metadata sheet in the Excel file. The model id is expected to be a tuple of (prefix, title)
            (space, external_id) for Conceptual and Physical Data Model respectively.

        sheet_prefix: The prefix to use for the sheet names in the Excel file. Defaults to an empty string.

    The following styles are available:

    - "none":    No styling is applied.
    - "minimal": Column widths are adjusted to fit the content, and the header row(s) is frozen.
    - "default": Minimal + headers are bold, increased size, and colored.
    - "maximal": Default + alternating row colors in the properties sheet for each class in addition to extra
                 blank rows between classes and borders
    """

    Style = Literal["none", "minimal", "default", "maximal"]
    DumpOptions = Literal["user", "last", "reference"]
    _helper_sheet_name: str = "_helper"
    _main_header_by_sheet_name: ClassVar[dict[str, str]] = {
        "Properties": "Definition of Properties",
        "Concepts": "Definition of Concepts",
        "Views": "Definition of Views",
        "Containers": "Definition of Containers",
        "Nodes": "Definition of Nodes",
        "Enum": "Definition of Enum Collections",
    }
    _helper_sheet_column_indexes_by_names: ClassVar[dict[str, int]] = {
        "Concept": 1,
        "View": 1,
        "Implements": 2,
        "Value Type": 3,
        "Container": 4,
        "In Model": 5,
        "Immutable": 5,
        "Used For": 6,
    }
    style_options = get_args(Style)
    dump_options = get_args(DumpOptions)

    def __init__(
        self,
        styling: Style = "default",
        new_model_id: tuple[str, str] | None = None,
        sheet_prefix: str | None = None,
        reference_data_model_with_prefix: tuple[VerifiedDataModel, str] | None = None,
        add_empty_rows: bool = False,
        hide_internal_columns: bool = True,
        include_properties: Literal["same-space", "all"] = "all",
        add_drop_downs: bool = True,
        base_model: BASE_MODEL | None = None,
        total_concepts: int | None = None,
    ):
        self.sheet_prefix = sheet_prefix or ""
        if styling not in self.style_options:
            raise ValueError(f"Invalid styling: {styling}. Valid options are {self.style_options}")
        self.styling = styling
        self._styling_level = self.style_options.index(styling)
        self.new_model_id = new_model_id
        self.reference_data_model_with_prefix = reference_data_model_with_prefix
        self.add_empty_rows = add_empty_rows
        self.hide_internal_columns = hide_internal_columns
        self.include_properties = include_properties
        self.add_drop_downs = add_drop_downs
        self.base_model = base_model
        self.total_concepts = total_concepts

    @property
    def description(self) -> str:
        return "Export verified model to Excel."

    def export_to_file(self, data_model: VerifiedDataModel, filepath: Path) -> None:
        """Exports transformation data_model to excel file."""
        data = self.export(data_model)
        try:
            data.save(filepath)
        finally:
            data.close()
        return None

    @overload
    def template(self, role: RoleTypes, filepath: Path) -> None: ...

    @overload
    def template(self, role: RoleTypes, filepath: Path | None = None) -> None: ...

    def template(self, role: RoleTypes, filepath: Path | None = None) -> None | Workbook:
        """This method will create an spreadsheet template for data modeling depending on the role.

        Args:
            role: The role for which the template is created. Can be either "dms" or "information".
            filepath: The path to the file where the template will be saved.

        """
        workbook = Workbook()
        # Remove default sheet named "Sheet"
        workbook.remove(workbook["Sheet"])

        data_model_model = PhysicalDataModel if role == RoleTypes.dms else ConceptualDataModel

        headers_by_sheet = data_model_model.headers_by_sheet(by_alias=True)
        headers_by_sheet.pop("Metadata")

        self._write_metadata_sheet(
            workbook,
            cast(BaseVerifiedMetadata, data_model_model.model_fields["metadata"].annotation).default().model_dump(),
        )

        for sheet_name, headers in headers_by_sheet.items():
            if sheet_name in ("Metadata", "Prefixes", "Reference", "Last"):
                continue
            sheet = self._create_sheet_with_header(workbook, headers, sheet_name)
            self._style_sheet_header(sheet, headers)

        self._adjust_column_widths(workbook)
        self._hide_internal_columns(workbook)

        self._add_drop_downs(role, workbook)

        if filepath:
            try:
                workbook.save(filepath)
            finally:
                workbook.close()
            return None

        return workbook

    def export(self, data_model: VerifiedDataModel) -> Workbook:
        workbook = Workbook()
        # Remove default sheet named "Sheet"
        workbook.remove(workbook["Sheet"])

        dumped_user_data_model: dict[str, Any] = data_model.dump(by_alias=True)

        self._write_metadata_sheet(workbook, dumped_user_data_model["Metadata"], sheet_prefix=self.sheet_prefix)
        self._write_sheets(workbook, dumped_user_data_model, data_model, sheet_prefix=self.sheet_prefix)
        if self.reference_data_model_with_prefix:
            reference_data_model, prefix = self.reference_data_model_with_prefix
            dumped_reference_data_model = reference_data_model.dump(entities_exclude_defaults=False, by_alias=True)
            self._write_sheets(workbook, dumped_reference_data_model, reference_data_model, sheet_prefix=prefix)
            self._write_metadata_sheet(workbook, dumped_reference_data_model["Metadata"], sheet_prefix=prefix)

        if isinstance(data_model, ConceptualDataModel) and data_model.prefixes:
            self._write_prefixes_sheet(workbook, data_model.prefixes)

        if self._styling_level > 0:
            self._adjust_column_widths(workbook)

        if self.hide_internal_columns:
            self._hide_internal_columns(workbook)

        # Only add drop downs if the data_model are Physical Data Model
        if self.add_drop_downs:
            self._add_drop_downs(data_model.metadata.role, workbook)

        return workbook

    def _hide_internal_columns(self, workbook: Workbook) -> None:
        """Hides internal columns in workbook sheets.

        Args:
            workbook: Workbook representation of the Excel file.

        """
        for sheet in workbook.sheetnames:
            if sheet.lower() == "metadata":
                continue
            ws = workbook[sheet]
            for col in get_internal_properties():
                column_letter = find_column_and_row_with_value(ws, col)[0]
                if column_letter:
                    ws.column_dimensions[column_letter].hidden = True

    def _add_drop_downs(
        self,
        role: RoleTypes,
        workbook: Workbook,
    ) -> None:
        """Adds drop down menus to specific columns for fast and accurate data entry
        for both Conceptual and Physical Data Model.

        Args:
            role: The role for which the drop downs are created. Can be either "dms" or "information".
            workbook: Workbook representation of the Excel file.

        !!! note "Why defining individual data validation per desired column?
            This is due to the internal working of openpyxl. Adding same validation to
            different column leads to unexpected behavior when the openpyxl workbook is exported
            as and Excel file. Probably, the validation is not copied to the new column,
            but instead reference to the data validation object is added.

        !!! note "Why 100*100 rows?"
            This is due to the fact that we need to add drop down menus for all properties
            in the sheet. The maximum number of properties per view/concept is 100. So considering
            maximum number of views/concepts is 100, we need to add 100*100 rows.
        """

        self._make_helper_sheet(role, workbook, 100)

        data_validators: dict[str, DataValidation] = {}

        # We need create individual data validation and cannot re-use the same one due
        # the internals of openpyxl

        self._add_data_validation(
            workbook,
            sheet_name="Views" if role == RoleTypes.dms else "Concepts",
            column_name="Implements",
            data_validator_name="implements",
            data_validators=data_validators,
            validation_range=100 + 100,  # base + user concepts (max)
            total_rows=100,
        )

        self._add_data_validation(
            workbook,
            sheet_name="Properties",
            column_name="Value Type",
            data_validator_name="value_type",
            data_validators=data_validators,
            validation_range=150,  # primitive types + classes
            total_rows=100 * 100,  # 100 views/classes * 100 properties (max properties per view/class)
        )

        self._add_data_validation(
            workbook,
            sheet_name="Properties",
            column_name="View" if role == RoleTypes.dms else "Concept",
            data_validator_name="views_or_concepts",
            data_validators=data_validators,
            validation_range=100,
            total_rows=100 * 100,
        )

        if role == RoleTypes.dms:
            self._add_data_validation(
                workbook,
                sheet_name="Properties",
                column_name="Container",
                data_validator_name="container",
                data_validators=data_validators,
                validation_range=100,
                total_rows=100 * 100,
            )

            self._add_data_validation(
                workbook,
                sheet_name="Properties",
                column_name="Immutable",
                data_validator_name="immutable",
                data_validators=data_validators,
                validation_range=2,
                total_rows=100 * 100,
            )

            self._add_data_validation(
                workbook,
                sheet_name="Views",
                column_name="In Model",
                data_validator_name="in_model",
                data_validators=data_validators,
                validation_range=2,
                total_rows=100,  # 100 views
            )

            self._add_data_validation(
                workbook,
                sheet_name="Containers",
                column_name="Used For",
                data_validator_name="used_for",
                data_validators=data_validators,
                validation_range=3,
                total_rows=100,  # 100 views
            )

    def _make_helper_sheet(
        self,
        role: RoleTypes,
        workbook: Workbook,
        no_rows: int,
    ) -> None:
        """This helper sheet is used as source of data for drop down menus creation

        Args:
            role: The role for which the helper sheet is created. Can be either "dms" or "information".
            workbook: Workbook representation of the Excel file.
            no_rows: number of rows to add data too that will form base for drop down menus.
        """

        workbook.create_sheet(title=self._helper_sheet_name)

        value_type_counter = 0

        for value_type_counter, value_type in enumerate(_DATA_TYPE_BY_DMS_TYPE.values()):
            value_type_as_str = value_type.dms._type.casefold() if role == RoleTypes.dms else value_type.xsd
            # skip types which require special handling or are surpassed by CDM
            if value_type_as_str in ["enum", "timeseries", "sequence", "file"]:
                continue
            workbook[self._helper_sheet_name].cell(
                row=value_type_counter + 1,
                column=self._helper_sheet_column_indexes_by_names["Value Type"],
                value=value_type_as_str,
            )

        value_type_counter += 1

        concept_counter = 0
        if self.base_model and (concepts := get_base_concepts(self.base_model, self.total_concepts)):
            for concept_counter, concept in enumerate(concepts):
                workbook[self._helper_sheet_name].cell(
                    row=concept_counter + 1,
                    column=self._helper_sheet_column_indexes_by_names["Implements"],
                    value=concept,
                )
            concept_counter += 1

        views_or_concepts_sheet = "Views" if role == RoleTypes.dms else "Concepts"
        view_or_concept_column = "View" if role == RoleTypes.dms else "Concept"

        for i in range(no_rows):
            workbook[self._helper_sheet_name].cell(
                row=i + 1,
                column=self._helper_sheet_column_indexes_by_names[view_or_concept_column],
                value=f'=IF(ISBLANK({views_or_concepts_sheet}!A{i + 3}), "", {views_or_concepts_sheet}!A{i + 3})',
            )

            workbook[self._helper_sheet_name].cell(
                row=concept_counter + i + 1,
                column=self._helper_sheet_column_indexes_by_names["Implements"],
                value=f'=IF(ISBLANK({views_or_concepts_sheet}!A{i + 3}), "", {views_or_concepts_sheet}!A{i + 3})',
            )

            workbook[self._helper_sheet_name].cell(
                row=value_type_counter + i + 1,
                column=self._helper_sheet_column_indexes_by_names["Value Type"],
                value=f'=IF(ISBLANK({views_or_concepts_sheet}!A{i + 3}), "", {views_or_concepts_sheet}!A{i + 3})',
            )

            if role == RoleTypes.dms:
                workbook[self._helper_sheet_name].cell(
                    row=i + 1,
                    column=self._helper_sheet_column_indexes_by_names["Container"],
                    value=f'=IF(ISBLANK(Containers!A{i + 3}), "", Containers!A{i + 3})',
                )

        if role == RoleTypes.dms:
            for i, value in enumerate([True, False, ""]):
                workbook[self._helper_sheet_name].cell(
                    row=i + 1,
                    column=self._helper_sheet_column_indexes_by_names["In Model"],
                    value=cast(bool | str, value),
                )

            for i, value in enumerate(["node", "edge", "all"]):
                workbook[self._helper_sheet_name].cell(
                    row=i + 1,
                    column=self._helper_sheet_column_indexes_by_names["Used For"],
                    value=value,
                )

        workbook[self._helper_sheet_name].sheet_state = "hidden"

    def _add_data_validation(
        self,
        workbook: Workbook,
        sheet_name: str,
        column_name: str,
        data_validator_name: str,
        data_validators: dict,
        validation_range: int,
        total_rows: int,
    ) -> None:
        """Adds data validation to a column in a sheet.

        Args:
            workbook: Workbook representation of the Excel file.
            sheet_name: The name of the sheet to add the data validation to.
            column_name: The name of the column to add the data validation to.
            data_validator_name: The name of the data validation to add.
            data_validators: A dictionary to store the data validators.
            validation_range: The total number of validation values to add.
            total_rows: The number of rows to add the data validation to.

        !!! note "Why defining individual data validation per desired column?"
            This is due to the internal working of openpyxl. Adding same validation to
            different column leads to unexpected behavior when the openpyxl workbook is exported
            as and Excel file.

        !!! note "Why starting at row 3?"
            This is due to the header rows in the sheet. The first two rows are reserved for the header.
        """
        # CREATE VALIDATOR
        data_validators[data_validator_name] = generate_data_validation(
            self._helper_sheet_name,
            get_column_letter(self._helper_sheet_column_indexes_by_names[column_name]),
            total_header_rows=0,
            validation_range=validation_range,
        )

        # REGISTER VALIDATOR TO SPECIFIC WORKBOOK SHEET
        workbook[sheet_name].add_data_validation(data_validators[data_validator_name])

        # APPLY VALIDATOR TO SPECIFIC COLUMN
        if column_letter := find_column_and_row_with_value(workbook[sheet_name], column_name)[0]:
            data_validators[data_validator_name].add(f"{column_letter}{3}:{column_letter}{3 + total_rows}")

    def _create_sheet_with_header(
        self,
        workbook: Workbook,
        headers: list[str],
        sheet_name: str,
        sheet_prefix: str = "",
    ) -> Worksheet:
        """Creates an empty sheet with the given headers.

        Args:
            workbook: The workbook to add the sheet to.
            headers: The headers to add to the sheet.
            sheet_name: The name of the sheet.
            sheet_prefix: The prefix to add to the sheet name, if any.
        """

        sheet = workbook.create_sheet(f"{sheet_prefix}{sheet_name}")
        main_header = self._main_header_by_sheet_name[sheet_name]
        sheet.append([main_header] + [""] * (len(headers) - 1))

        if headers[0] == "Neat ID":
            # Move the Neat ID to the end of the columns
            headers = headers[1:] + ["Neat ID"]

        # Append the headers to the sheet
        sheet.append(headers)

        return sheet

    def _style_sheet_header(self, sheet: Worksheet, headers: list[str]) -> None:
        """Styles the sheet with the given headers.

        Args:
            sheet: The sheet to style.
            headers: The headers to style.
        """
        if self._styling_level > 0:
            # This freezes all rows above the given row
            sheet.freeze_panes = sheet["A3"]

            sheet["A1"].alignment = Alignment(horizontal="left")

        if self._styling_level > 1:
            # Make the header row bold, larger, and colored
            for cell, *_ in sheet.iter_cols(min_row=1, max_row=1, min_col=1, max_col=len(headers)):
                cell.font = Font(bold=True, size=20)
                cell.fill = PatternFill(fgColor="FFC000", patternType="solid")
            for cell in sheet["2"]:
                cell.font = Font(bold=True, size=14)

    def _write_sheets(
        self,
        workbook: Workbook,
        dumped_data_model: dict[str, Any],
        data_model: VerifiedDataModel,
        sheet_prefix: str = "",
    ) -> None:
        for sheet_name, headers in data_model.headers_by_sheet(by_alias=True).items():
            if sheet_name in ("Metadata", "Prefixes", "Reference", "Last"):
                continue

            sheet = self._create_sheet_with_header(workbook, headers, sheet_name, sheet_prefix)

            fill_colors = itertools.cycle(["CADCFC", "FFFFFF"])
            fill_color = next(fill_colors)
            last_concept: str | None = None
            item: dict[str, Any]
            for item in dumped_data_model.get(sheet_name) or []:
                if "Neat ID" in item:
                    # Move the Neat ID to the end of the columns
                    item["Neat ID"] = item.pop("Neat ID")
                row = list(item.values())
                concept = row[0]

                is_properties = sheet_name == "Properties"
                is_new_class = concept != last_concept and last_concept is not None
                if self._styling_level > 2 and is_new_class and is_properties:
                    if self.add_empty_rows:
                        sheet.append([""] * len(headers))
                    for cell in sheet[sheet.max_row]:
                        cell.fill = PatternFill(fgColor=fill_color, patternType="solid")
                        side = Side(style="thin", color="000000")
                        cell.border = Border(left=side, right=side, top=side, bottom=side)
                    fill_color = next(fill_colors)

                if is_properties and self.include_properties == "same-space":
                    space = concept.split(":")[0] if ":" in concept else data_model.metadata.space
                    if space != data_model.metadata.space:
                        continue

                sheet.append(row)
                if self._styling_level > 2 and is_properties:
                    for cell in sheet[sheet.max_row]:
                        cell.fill = PatternFill(fgColor=fill_color, patternType="solid")
                        side = Side(style="thin", color="000000")
                        cell.border = Border(left=side, right=side, top=side, bottom=side)
                last_concept = concept

            self._style_sheet_header(sheet, headers)

    def _write_metadata_sheet(self, workbook: Workbook, metadata: dict[str, Any], sheet_prefix: str = "") -> None:
        # Excel does not support timezone in datetime strings
        if isinstance(metadata.get("created"), datetime):
            metadata["created"] = metadata["created"].replace(tzinfo=None)
        if isinstance(metadata.get("updated"), datetime):
            metadata["updated"] = metadata["updated"].replace(tzinfo=None)

        metadata_sheet = workbook.create_sheet(f"{sheet_prefix}Metadata")
        for key, value in metadata.items():
            metadata_sheet.append([key, value])

        if self._styling_level > 1:
            for cell in metadata_sheet["A"]:
                cell.font = Font(bold=True, size=12)

    def _write_prefixes_sheet(self, workbook: Workbook, prefixes: dict[str, Namespace]) -> None:
        metadata_sheet = workbook.create_sheet("Prefixes")
        metadata_sheet.append(["Prefix", "Namespace"])
        for key, value in prefixes.items():
            metadata_sheet.append([key, value])

        if self._styling_level > 1:
            for cell in metadata_sheet["A"]:
                cell.font = Font(bold=True, size=12)

    @classmethod
    def _get_item_class(cls, annotation: GenericAlias) -> type[SheetRow]:
        if not isinstance(annotation, GenericAlias):
            raise ValueError(f"Expected annotation to be a GenericAlias, but got {type(annotation)}")
        args = get_args(annotation)
        if len(args) != 1:
            raise ValueError(f"Expected annotation to have exactly one argument, but got {len(args)}")
        arg = args[0]
        if not issubclass(arg, SheetRow):
            raise ValueError(f"Expected annotation to have a BaseModel argument, but got {type(arg)}")
        return arg

    @classmethod
    def _adjust_column_widths(cls, workbook: Workbook) -> None:
        for sheet_ in workbook:
            sheet = cast(Worksheet, sheet_)
            for column_cells in sheet.columns:
                try:
                    max_length = max(len(str(cell.value)) for cell in column_cells if cell.value is not None)
                except ValueError:
                    max_length = 0

                selected_column = column_cells[0]
                if isinstance(selected_column, MergedCell):
                    selected_column = column_cells[1]

                current = sheet.column_dimensions[selected_column.column_letter].width or (max_length + 0.5)  # type: ignore[union-attr]
                sheet.column_dimensions[selected_column.column_letter].width = min(  # type: ignore[union-attr]
                    max(current, max_length + 0.5), MAX_COLUMN_WIDTH
                )
        return None



from pydantic import BaseModel, Field
from pydantic.alias_generators import to_camel

from cognite.neat._data_model._identifiers import URI


class BaseModelObject(BaseModel, alias_generator=to_camel, extra="ignore", populate_by_name=True):
    """Base class for all object. This includes resources and nested objects."""

    ...


class ResourceMetadata(BaseModelObject):
    name: str | None = Field(
        None, description="Human readable / display name of resource being described.", max_length=1024
    )
    description: str | None = Field(None, description="The description of the resource.", max_length=255)
    uri: URI | None = Field(None, description="The URI of the resource being described.")



import json

from cognite.client import ClientConfig, CogniteClient

from cognite.neat import _version
from cognite.neat._client import NeatClient
from cognite.neat._config import NeatConfig, PredefinedProfile
from cognite.neat._session._usage_analytics._collector import Collector
from cognite.neat._state_machine import EmptyState, PhysicalState
from cognite.neat._store import NeatStore
from cognite.neat._utils.http_client import ParametersRequest, SuccessResponse

from ._cdf import CDF
from ._issues import Issues
from ._physical import PhysicalDataModel
from ._result import Result


def _is_in_browser() -> bool:
    try:
        from pyodide.ffi import IN_BROWSER  # type: ignore [import-not-found]
    except ModuleNotFoundError:
        return False
    return IN_BROWSER


class NeatSession:
    """A session is an interface for neat operations."""

    def __init__(
        self, client: CogniteClient | ClientConfig | None, config: PredefinedProfile | NeatConfig = "legacy-additive"
    ) -> None:
        """Initialize a Neat session.

        Args:
            client (CogniteClient | ClientConfig): The Cognite client or client configuration to use for the session.
            config (Literal["legacy-additive", "legacy-rebuild", "deep-additive", "deep-rebuild"] | NeatConfig):
                The configuration profile to use for the session.
                Defaults to "legacy-additive". This means Neat will perform additive modeling
                and apply only validations that were part of the legacy Neat version.
        """
        if client is None and _is_in_browser():
            client = CogniteClient()
        elif client is None:
            raise ValueError("A CogniteClient or ClientConfig must be provided to initialize a NeatSession.")
        self._config = NeatConfig.create_predefined(config) if isinstance(config, str) else config

        # Use configuration for physical data model
        self._client = NeatClient(client)
        self._store = NeatStore(config=self._config, client=self._client)
        self.physical_data_model = PhysicalDataModel(self._store, self._client, self._config)
        self.issues = Issues(self._store)
        self.result = Result(self._store)

        if self._config.alpha.enable_cdf_analysis:
            self.cdf = CDF(self._store, self._client, self._config)

        collector = Collector()
        if collector.can_collect:
            collector.collect("initSession", {"mode": self._config.modeling.mode})

        self._welcome_message()

    def _welcome_message(self) -> None:
        cdf_project = self._client.config.project
        message = f"Neat session started for CDF project: '{cdf_project}'"
        responses = self._client.http_client.request(
            ParametersRequest(endpoint_url=self._client.config.create_api_url(""), method="GET")
        )
        if len(responses) == 1 and isinstance(response := responses[0], SuccessResponse):
            organization = ""
            try:
                organization = json.loads(response.body)["organization"]
            except (KeyError, ValueError):
                ...
            if organization:
                message += f" (Organization: '{organization}')"

        print(message)
        print(self._config)

    @property
    def version(self) -> str:
        """Get the current version of neat."""
        return _version.__version__

    def _repr_html_(self) -> str:
        if isinstance(self._store.state, EmptyState):
            return (
                "<strong>Empty session</strong>. Get started by reading for example physical data model"
                " <em>.physical_data_model.read</em>"
            )

        if isinstance(self._store.state, PhysicalState):
            return self.physical_data_model._repr_html_()

        raise RuntimeError("Unknown session state, contact support.")
